import argparse
import asyncio
import base64
import io
import json
import os
import random
import time
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import Any, AsyncGenerator, Collection, Dict, List, Optional, Tuple

import numpy as np
from .backend_request_func import (ASYNC_REQUEST_FUNCS, RequestFuncInput,
                                  RequestFuncOutput)
from datasets import load_dataset
from PIL.Image import Image
from tqdm.asyncio import tqdm

from anc.data.jsonl_dataset import JsonlDataset
from anc.data.anc_processor import Processor  
from .multidialog_processor import MultiDialogProcessor
from .metric_collector import MetricCollector

try:
    from vllm.transformers_utils.tokenizer import get_tokenizer
except ImportError:
    from .backend_request_func import get_tokenizer

try:
    from vllm.utils import FlexibleArgumentParser
except ImportError:
    from argparse import ArgumentParser as FlexibleArgumentParser

MILLISECONDS_TO_SECONDS_CONVERSION = 1000

from .server import ServerFactory 
class LoadTestingOperator:
    def __init__(self):
        self.server = None
        self.metric_collector = None
    def start_server(self, dry_run, server_log_dir, start_otlp_server=False, **kwargs):
        """
        Start a server for load testing based on the specified backend.
        
        Args:
            **kwargs: Arguments for server initialization, must include 'backend'
        
        Returns:
            bool: True if server started successfully, False otherwise
        """
        # If args is a Namespace, convert to dict
        if hasattr(kwargs, '__dict__'):
            kwargs = vars(kwargs)
            
        # Extract backend type from kwargs or use default
        backend = kwargs.get('backend', 'vllm')
        try:
            if self.server:
                self.server.stop()
            #start vllm server
            self.server = ServerFactory.create_server(backend, **kwargs)
            self.server.start(dry_run=dry_run, server_log_dir=server_log_dir, start_otlp_server=start_otlp_server)
            # Start the metric collector
            if start_otlp_server:
                self.start_metric_collector(server_log_dir)
        except Exception as e:
            print(f"Failed to start {backend} server: {e}")
            raise e 

    def start_metric_collector(self, server_log_dir):
        server_metric_path = os.path.join(server_log_dir, f"otlp_metric_{int(time.time())}.json")
        self.metric_collector = MetricCollector(export_file_path=server_metric_path)
        self.metric_collector.start()
        print(f"\n{'-'*100}")
        print(f"Metric collector will save data to {server_metric_path}")
        print(f"{'-'*100}\n")
             
    def stop_server(self):
        self.server.stop()
        # Stop the metric collector
        if self.metric_collector:
            self.metric_collector.stop()

    def load_test(self, args: argparse.Namespace):
        """
        Start a load test with the given parameters.
        
        :param test_plan: Path to the test plan file
        :param duration: Duration of the load test in seconds
        :param concurrency: Number of concurrent users
        """
        print(args)
        if args.seed is not None:
            random.seed(args.seed)
            np.random.seed(args.seed)
        max_concurrency = args.max_concurrency

        backend = args.backend
        model_id = args.model_id if args.model_id is not None else args.model
        tokenizer_id = args.tokenizer if args.tokenizer is not None else args.model

        if args.base_url is not None:
            api_url = f"{args.base_url}{args.endpoint}"
            base_url = f"{args.base_url}"
        else:
            api_url = f"http://{args.host}:{args.port}{args.endpoint}"
            base_url = f"http://{args.host}:{args.port}"

        tokenizer = get_tokenizer(tokenizer_id,
                                trust_remote_code=args.trust_remote_code)


        if args.dataset_name == "anc":
            if args.backend == "vllm":
                input_requests = sample_anc_requests(
                    dataset_path=args.dataset_path,
                    num_requests=args.num_prompts,
                    tokenizer=tokenizer,
                    processor_class=MultiDialogProcessor,
                    fixed_output_len=args.anc_output_len,
                    multi_turn=args.multi_turn
                )
            elif args.backend == "openai-chat":
                input_requests = sample_anc_requests_openai_chat(
                    dataset_path=args.dataset_path,
                    num_requests=args.num_prompts,
                    fixed_output_len=args.anc_output_len
                )
            else:
                raise ValueError(f"backend: {args.backend} is not supported for anc dataset yet")

        elif args.dataset_name == "hf":
            input_requests = sample_hf_requests(
                dataset_path=args.dataset_path,
                dataset_subset=args.hf_subset,
                dataset_split=args.hf_split,
                num_requests=args.num_prompts,
                tokenizer=tokenizer,
                random_seed=args.seed,
                fixed_output_len=args.hf_output_len,
            )

        elif args.dataset_name == "random":
            input_requests = sample_random_requests(
                prefix_len=args.random_prefix_len,
                input_len=args.random_input_len,
                output_len=args.random_output_len,
                num_prompts=args.num_prompts,
                range_ratio=args.random_range_ratio,
                tokenizer=tokenizer,
            )

        else:
            raise ValueError(f"Unknown dataset: {args.dataset_name}")

        gootput_config_dict = check_goodput_args(args)
        try:
            benchmark_result = asyncio.run(
            benchmark(
                backend=backend,
                api_url=api_url,
                base_url=base_url,
                model_id=model_id,
                tokenizer=tokenizer,
                input_requests=input_requests,
                logprobs=args.logprobs,
                best_of=args.best_of,
                request_rate=args.request_rate,
                burstiness=args.burstiness,
                disable_tqdm=args.disable_tqdm,
                profile=args.profile,
                selected_percentile_metrics=args.percentile_metrics.split(","),
                selected_percentiles=[
                    float(p) for p in args.metric_percentiles.split(",")
                ],
                ignore_eos=args.ignore_eos,
                gootput_config_dict=gootput_config_dict,
                max_concurrency=max_concurrency,
                    num_warmup_requests=int(args.num_warmup_requests),
                ))
            self.process_and_save_results(args, max_concurrency, backend, model_id, tokenizer_id, benchmark_result)

            try:
                print(f'\n{"-"*100}')
                print(f'Collecting prometheus metrics')
                print(f"\n{'-'*100}")
                from .prometheus_metric import collect_prometheus_metrics, analyze_metrics 
                current_time = int(time.time())
                raw_prometheus_metrics_path = os.path.join(args.result_dir, f"raw_prometheus_metrics_{current_time}.txt")
                processed_prometheus_metrics_path = os.path.join(args.result_dir, f"processed_prometheus_metrics_{current_time}.json")
                raw_prometheus_metrics = collect_prometheus_metrics(server_url=f"{base_url}", save_raw_path=raw_prometheus_metrics_path)
                analyze_metrics(raw_prometheus_metrics, save_results_path=processed_prometheus_metrics_path)
                print(f'Prometheus metrics saved to {processed_prometheus_metrics_path}')
            except Exception as e:
                print(f"Warning: Failed to collect prometheus metrics: {e}")

            if self.metric_collector:
                print(f"\n{'-'*100}")
                print(f"Waiting for metric collector to flush data")
                print(f"\n{'-'*100}")
                time.sleep(10) # wait for metric collector to flush data
                self.metric_collector._analyze()
        except Exception as e:
            print(f"Failed to run benchmark: {e}")
            raise e
        finally:
            # keep the metric collector running if user wants to keep vllm server running. for debugging purpose
            if self.metric_collector and not args.skip_server:
                self.metric_collector.stop()
        # Save config and results to json

        return benchmark_result

        def stop_test(self, test_id):
            """
            Stop a load test with the given test ID.
            
            :param test_id: ID of the test to stop
            """
            # Implement the logic to stop the load test
            print(f"Stopping load test with ID {test_id}.")
            # Add your load testing logic here

        def get_test_status(self, test_id):
            """
            Get the status of a load test with the given test ID.
            
            :param test_id: ID of the test to get status
            """
            # Implement the logic to get the status of the load test
            print(f"Getting status for load test with ID {test_id}.")

    def process_and_save_results(self, args, max_concurrency, backend, model_id, tokenizer_id, benchmark_result):
        if args.save_result:
            result_json: Dict[str, Any] = {}

            # Setup
            current_dt = datetime.now().strftime("%Y%m%d-%H%M%S")
            result_json["date"] = current_dt
            result_json["backend"] = backend
            result_json["model_id"] = model_id
            result_json["tokenizer_id"] = tokenizer_id
            result_json["best_of"] = args.best_of
            result_json["num_prompts"] = args.num_prompts

            # Metadata
            if args.metadata:
                for item in args.metadata:
                    if "=" in item:
                        kvstring = item.split("=")
                        result_json[kvstring[0].strip()] = kvstring[1].strip()
                    else:
                        raise ValueError(
                            "Invalid metadata format. Please use KEY=VALUE format."
                        )

            # Traffic
            result_json["request_rate"] = (
                args.request_rate if args.request_rate < float("inf") else "inf")
            result_json["burstiness"] = args.burstiness
            result_json["max_concurrency"] = max_concurrency
            

            # Merge with benchmark result
            result_json = { **benchmark_result,**result_json}

            # Save to file
            base_model_id = model_id.split("/")[-1]
            max_concurrency_str = (f"-concurrency{max_concurrency}"
                                if max_concurrency is not None else "")
            file_name = f"{backend}-{args.request_rate}qps{max_concurrency_str}-{base_model_id}-{current_dt}.json"  #noqa
            if args.result_filename:
                file_name = args.result_filename
            if args.result_dir:
                file_name = os.path.join(args.result_dir, file_name)
            with open(file_name, "w", encoding='utf-8') as outfile:
                json.dump(result_json, outfile)
            # Add your load testing logic here




@dataclass
class BenchmarkMetrics:
    completed: int
    total_input: int
    total_output: int
    request_throughput: float
    request_goodput: float
    output_throughput: float
    total_token_throughput: float
    mean_input_token_len: int
    median_input_token_len: int
    std_input_token_len: int
    percentiles_input_token_len: List[Tuple[float, int]]
    mean_output_token_len: int
    median_output_token_len: int
    std_output_token_len: int
    percentiles_output_token_len: List[Tuple[float, int]]
    mean_fs_token_len: int
    median_fs_token_len: int
    std_fs_token_len: int
    percentiles_fs_token_len: List[Tuple[float, int]]
    percentiles_input_token_len: List[Tuple[float, int]]
    mean_ttft_ms: float
    median_ttft_ms: float
    std_ttft_ms: float
    percentiles_ttft_ms: List[Tuple[float, float]]
    mean_ttfs_ms: float
    median_ttfs_ms: float
    std_ttfs_ms: float
    percentiles_ttfs_ms: List[Tuple[float, float]]
    mean_tpot_ms: float
    median_tpot_ms: float
    std_tpot_ms: float
    percentiles_tpot_ms: List[Tuple[float, float]]
    mean_itl_ms: float
    median_itl_ms: float
    std_itl_ms: float
    percentiles_itl_ms: List[Tuple[float, float]]
    # E2EL stands for end-to-end latency per request.
    # It is the time taken on the client side from sending
    # a request to receiving a complete response.
    mean_e2el_ms: float
    median_e2el_ms: float
    std_e2el_ms: float
    percentiles_e2el_ms: List[Tuple[float, float]]


def sample_mmmu_pro_vision_requests(
    dataset,
    num_requests: int,
    tokenizer,
    fixed_output_len: Optional[int] = None,
) -> List[Tuple[str, str, int, Optional[Dict[str, Collection[str]]]]]:
    sampled_requests: List[Tuple[str, int, int, Dict[str,
                                                     Collection[str]]]] = []
    for data in dataset:
        if len(sampled_requests) == num_requests:
            break

        # MMMU-Pro vision direct prompt
        # Ref: https://github.com/MMMU-Benchmark/MMMU/blob/6ce42f4d8f70c1841c67867152648974415b5cac/mmmu-pro/prompts.yaml#L5
        prompt = (
            "Answer with the option letter from the given choices directly. "
            "The last line of your response should be of the following "
            "format: 'Answer: $LETTER' (without quotes) where LETTER is one of "
            "options.")

        prompt_token_ids = tokenizer(prompt).input_ids
        if fixed_output_len is None:
            # Default max output len is set to 128
            print("--hf-output-len is not provided. Using default value 128.")
            fixed_output_len = 128

        prompt_len = len(prompt_token_ids)
        output_len = fixed_output_len

        assert isinstance(
            data["image"],
            Image), ("Input image format must be `PIL.Image.Image`, "
                     f"given {type(data['image'])}.")
        image: Image = data["image"]
        image = image.convert("RGB")
        image_data = io.BytesIO()
        image.save(image_data, format='JPEG')
        image_base64 = base64.b64encode(image_data.getvalue()).decode("utf-8")
        mm_content = {
            "type": "image_url",
            "image_url": {
                "url": f"data:image/jpeg;base64,{image_base64}"
            },
        }

        sampled_requests.append((prompt, prompt_len, output_len, mm_content))

    return sampled_requests


def sample_anc_requests_openai_chat(
    dataset_path: str,
    num_requests: int,
    fixed_output_len: int,
) -> List[Tuple[str, int, int, Dict]]:
    supported_formats = ['.jsonl']
    _, file_extension = os.path.splitext(dataset_path)
    if not file_extension.lower() in supported_formats:
        raise ValueError(f"Unsupported dataset format: {dataset_path}. Supported Format: {supported_formats}")
    if file_extension == '.jsonl':
        # Initialize dataset and processor
        dataset = JsonlDataset(dataset_path)

        selected_ids = random.choices(range(len(dataset)), k=num_requests)
        print(f'selected_ids: {selected_ids}')
        processed_data = []
        for id in selected_ids:

            messages = dataset[id]['messages']
            if messages[-1]['role'] == 'assistant':
                messages.pop()
                
            processed_data.append([messages, None, fixed_output_len, None])

    return processed_data

def sample_anc_requests(
    dataset_path: str,
    num_requests: int,
    tokenizer,
    processor_class: Processor = None,
    fixed_output_len: Optional[int] = None,
    multi_turn: bool = False,
) -> List[Tuple[str, int, int, Dict]]:
    supported_formats = ['.jsonl']
    _, file_extension = os.path.splitext(dataset_path)
    if not file_extension.lower() in supported_formats:
        raise ValueError(f"Unsupported dataset format: {dataset_path}. Supported Format: {supported_formats}")
    
    if file_extension == '.jsonl':
        # Initialize dataset and processor
        dataset = JsonlDataset(dataset_path)
        processor = processor_class() if processor_class else MultiDialogProcessor()
        
        # Process and organize conversations
        conversation_groups = {}
        
        for idx in range(len(dataset)):
            item = dataset[idx]
            if idx==0:
                print(f'item: {item}')
            
            if isinstance(item, dict):
                if 'messages' in item:
                    messages = item['messages']
                else:
                    raise ValueError(f"Unsupported dataset format: {dataset_path}")
            else:
                messages = item
            
            transformed = processor.transform(messages)
            for data in transformed:
                prompt = data.get("prompt")
                output_len = data.get("output_len", fixed_output_len)
                conversation_id = data.get("conversation_id")
                turn_index = data.get("turn_index", 0)

                prompt_token_ids = tokenizer(prompt).input_ids
                prompt_len = len(prompt_token_ids)
                
                # Group by conversation ID
                if conversation_id not in conversation_groups:
                    conversation_groups[conversation_id] = []
                    
                # Store with metadata
                conversation_groups[conversation_id].append({
                    "prompt": prompt,
                    "prompt_len": prompt_len,
                    "output_len": output_len,
                    "turn_index": turn_index,
                    "conversation_id": conversation_id
                })
        
        # Sort each conversation by turn index
        for conv_id in conversation_groups:
            conversation_groups[conv_id].sort(key=lambda x: x["turn_index"])
        
        # Select a subset of conversations if needed
        if multi_turn:
            # we will send each prmopt in multi-turn conversations in order. num_requests is the total number of conversations
            conv_ids = list(conversation_groups.keys())
            selected_ids = random.choices(conv_ids, k=num_requests)
            conversation_groups = {conv_id: conversation_groups[conv_id] 
                                  for conv_id in selected_ids}
        
        # Flatten for processing but keep metadata

            processed_data = flatten_data(conversation_groups)

        else:
            # we will randomly pickup a single prompt in a random conversations, num_requests is the total number of prompts
            processed_data = flatten_data(conversation_groups, only_use_last_turn=True)
            selected_ids = random.choices(range(len(processed_data)), k=num_requests)
            processed_data = [processed_data[i] for i in selected_ids]
    
    print(f'total number of prompts: {len(processed_data)}')
    
    return processed_data

def flatten_data(conversation_groups, only_use_last_turn: bool = False):
    processed_data = []
    for conv_id, turns in conversation_groups.items():
        if only_use_last_turn:
            turns = turns[-1:]
        for turn in turns:
            processed_data.append((
                        turn["prompt"],
                        turn["prompt_len"],
                        turn["output_len"],
                        None
                    ))
    return processed_data

def sample_hf_requests(
    dataset_path: str,
    dataset_subset: str,
    dataset_split: str,
    num_requests: int,
    tokenizer,
    random_seed: int,
    fixed_output_len: Optional[int] = None,
) -> List[Tuple[str, str, int, Optional[Dict[str, Collection[str]]]]]:

    # Special case for MMMU-Pro vision dataset
    if dataset_path == 'MMMU/MMMU_Pro' and dataset_subset == 'vision':
        assert dataset_split == "test"
        dataset = load_dataset(dataset_path,
                               name=dataset_subset,
                               split=dataset_split,
                               streaming=True)
        assert "image" in dataset.features, (
            "MMMU/MMMU_Pro vision dataset must have 'image' column.")
        filter_func = lambda x: isinstance(x["image"], Image)
        dataset = dataset.shuffle(seed=random_seed).filter(filter_func)
        return sample_mmmu_pro_vision_requests(dataset, num_requests,
                                               tokenizer, fixed_output_len)

    dataset = load_dataset(dataset_path,
                           name=dataset_subset,
                           split=dataset_split,
                           streaming=True)
    assert "conversations" in dataset.features, (
        "HF Dataset must have 'conversations' column.")
    filter_func = lambda x: len(x["conversations"]) >= 2
    filtered_dataset = dataset.shuffle(seed=random_seed).filter(filter_func)
    sampled_requests: List[Tuple[str, int, int, Dict[str,
                                                     Collection[str]]]] = []
    for data in filtered_dataset:
        if len(sampled_requests) == num_requests:
            break

        # Tokenize the prompts and completions.
        prompt = data["conversations"][0]["value"]
        prompt_token_ids = tokenizer(prompt).input_ids
        completion = data["conversations"][1]["value"]
        completion_token_ids = tokenizer(completion).input_ids
        prompt_len = len(prompt_token_ids)
        output_len = len(completion_token_ids
                         ) if fixed_output_len is None else fixed_output_len
        if fixed_output_len is None and (prompt_len < 4 or output_len < 4):
            # Prune too short sequences.
            continue
        if fixed_output_len is None and \
            (prompt_len > 1024 or prompt_len + output_len > 2048):
            # Prune too long sequences.
            continue

        if "image" in data and isinstance(data["image"], Image):
            image: Image = data["image"]
            image = image.convert("RGB")
            image_data = io.BytesIO()
            image.save(image_data, format='JPEG')
            image_base64 = base64.b64encode(
                image_data.getvalue()).decode("utf-8")
            mm_content = {
                "type": "image_url",
                "image_url": {
                    "url": f"data:image/jpeg;base64,{image_base64}"
                },
            }
        elif "image" in data and isinstance(data["image"], str):
            if (data["image"].startswith("http://") or \
                data["image"].startswith("file://")):
                image_url = data["image"]
            else:
                image_url = f"file://{data['image']}"

            mm_content = {
                "type": "image_url",
                "image_url": {
                    "url": image_url
                },
            }
        else:
            mm_content = None

        sampled_requests.append((prompt, prompt_len, output_len, mm_content))

    return sampled_requests


def sample_random_requests(
    prefix_len: int,
    input_len: int,
    output_len: int,
    num_prompts: int,
    range_ratio: float,
    tokenizer,
) -> List[Tuple[str, int, int]]:
    prefix_token_ids = np.random.randint(0,
                                         tokenizer.vocab_size,
                                         size=prefix_len).tolist()

    input_lens = np.random.randint(
        int(input_len * range_ratio),
        input_len + 1,
        size=num_prompts,
    )
    output_lens = np.random.randint(
        int(output_len * range_ratio),
        output_len + 1,
        size=num_prompts,
    )
    offsets = np.random.randint(0, tokenizer.vocab_size, size=num_prompts)
    input_requests = []
    for i in range(num_prompts):
        input_ids = prefix_token_ids + [(offsets[i] + i + j) % tokenizer.vocab_size for j in range(input_lens[i])] 
        # input_ids =[i for i in range(input_lens[i])]
        input_requests.append((input_ids, int(prefix_len + input_lens[i]),
                               int(output_lens[i]), None))

    return input_requests


async def get_request(
    input_requests: List[Tuple[str, int, int]],
    request_rate: float,
    burstiness: float = 1.0,
) -> AsyncGenerator[Tuple[str, int, int], None]:
    """
    Asynchronously generates requests at a specified rate 
    with OPTIONAL burstiness.
    
    Args:
        input_requests: 
            A list of input requests, each represented as a tuple.
        request_rate: 
            The rate at which requests are generated (requests/s).
        burstiness (optional): 
            The burstiness factor of the request generation. 
            Only takes effect when request_rate is not inf.
            Default value is 1, which follows a Poisson process.
            Otherwise, the request intervals follow a gamma distribution.
            A lower burstiness value (0 < burstiness < 1) results 
            in more bursty requests, while a higher burstiness value 
            (burstiness > 1) results in a more uniform arrival of requests.
    """
    input_requests = iter(input_requests)

    # Calculate scale parameter theta to maintain the desired request_rate.
    assert burstiness > 0, (
        f"A positive burstiness factor is expected, but given {burstiness}.")
    theta = 1.0 / (request_rate * burstiness)

    for request in input_requests:
        yield request

        if request_rate == float("inf"):
            # If the request rate is infinity, then we don't need to wait.
            continue

        # Sample the request interval from the gamma distribution.
        # If burstiness is 1, it follows exponential distribution.
        interval = np.random.gamma(shape=burstiness, scale=theta)
        # The next request will be sent after the interval.
        await asyncio.sleep(interval)


def calculate_metrics(
    input_requests: List[Tuple[str, int, int]],
    outputs: List[RequestFuncOutput],
    dur_s: float,
    tokenizer,
    selected_percentile_metrics: List[str],
    selected_percentiles: List[float],
    gootput_config_dict: Dict[str, float],
) -> Tuple[BenchmarkMetrics, List[int]]:
    input_lens: List[int] = []
    actual_output_lens: List[int] = []
    actual_input_lens: List[int] = []
    actual_first_sentence_lens: List[int] = []
    total_input = 0
    completed = 0
    good_completed = 0
    itls: List[float] = []
    tpots: List[float] = []
    all_tpots: List[float] = []
    ttfts: List[float] = []
    ttfss: List[float] = []
    e2els: List[float] = []
    for i in range(len(outputs)):
        if outputs[i].success:
            # We use the tokenizer to count the number of output tokens for all
            # serving backends instead of looking at len(outputs[i].itl) since
            # multiple output tokens may be bundled together
            # Note : this may inflate the output token count slightly
            if outputs[i].input_token_len:
                actual_input_lens.append(outputs[i].input_token_len) # currently only available for openai completions  
            else:
                actual_input_lens.append(input_requests[i][1])
            input_len = outputs[i].input_token_len if outputs[i].input_token_len else input_requests[i][1]
            if outputs[i].output_token_len:
                output_len = outputs[i].output_token_len
            else:
                print(f'warning, output_token_len is not found for {input_requests[i][0]}, will use tokenizer to count the number of output tokens')
                output_len = len(
                    tokenizer(outputs[i].generated_text,
                          add_special_tokens=False).input_ids)
            
            actual_input_lens.append(input_len)
            actual_output_lens.append(output_len)
            
             
            first_sentence_len = len(
                tokenizer(outputs[i].first_sentence_text,
                          add_special_tokens=False).input_ids)
            actual_first_sentence_lens.append(first_sentence_len)
            
            # TODO: figure out why sometimes input_len is None: for anc data, soemtimes outputs[i].input_token_len is still None
            total_input += input_len if input_len else 0
            tpot = 0
            if output_len > 1:
                tpot = (outputs[i].latency - outputs[i].ttft) / (output_len -
                                                                 1)
                tpots.append(tpot)
            # Note: if output_len <= 1, we regard tpot as 0 for goodput
            all_tpots.append(tpot)
            itls += outputs[i].itl
            ttfts.append(outputs[i].ttft)
            ttfss.append(outputs[i].ttfs)
            e2els.append(outputs[i].latency)
            completed += 1
        else:
            input_lens.append(0)
            actual_output_lens.append(0)
            actual_first_sentence_lens.append(0)

    if gootput_config_dict:
        valid_metrics = []
        slo_values = []

        if "ttft" in gootput_config_dict:
            valid_metrics.append(ttfts)
            slo_values.append(gootput_config_dict["ttft"] /
                              MILLISECONDS_TO_SECONDS_CONVERSION)
        if "ttfs" in gootput_config_dict:
            valid_metrics.append(ttfss)
            slo_values.append(gootput_config_dict["ttfs"] /
                              MILLISECONDS_TO_SECONDS_CONVERSION)
        if "tpot" in gootput_config_dict:
            valid_metrics.append(all_tpots)
            slo_values.append(gootput_config_dict["tpot"] /
                              MILLISECONDS_TO_SECONDS_CONVERSION)
        if "e2el" in gootput_config_dict:
            valid_metrics.append(e2els)
            slo_values.append(gootput_config_dict["e2el"] /
                              MILLISECONDS_TO_SECONDS_CONVERSION)

        for req_metric in zip(*valid_metrics):
            is_good_req = all([s >= r for s, r in zip(slo_values, req_metric)])
            if is_good_req:
                good_completed += 1

    if completed == 0:
        warnings.warn(
            "All requests failed. This is likely due to a misconfiguration "
            "on the benchmark arguments.",
            stacklevel=2)
    metrics = BenchmarkMetrics(
        completed=completed,
        total_input=total_input,
        total_output=sum(actual_output_lens),
        request_throughput=completed / dur_s,
        request_goodput=good_completed / dur_s,
        output_throughput=sum(actual_output_lens) / dur_s,
        total_token_throughput=(total_input + sum(actual_output_lens)) / dur_s,
        mean_input_token_len=np.mean(actual_input_lens or 0),
        median_input_token_len=np.median(actual_input_lens or 0),
        std_input_token_len=np.std(actual_input_lens or 0),
        percentiles_input_token_len=[(p, np.percentile(actual_input_lens or 0, p))
                             for p in selected_percentiles], 
        mean_output_token_len=np.mean(actual_output_lens or 0),
        median_output_token_len=np.median(actual_output_lens or 0),
        std_output_token_len=np.std(actual_output_lens or 0),
        percentiles_output_token_len=[(p, np.percentile(actual_output_lens or 0, p))
                             for p in selected_percentiles], 
        mean_fs_token_len=np.mean(actual_first_sentence_lens or 0),
        median_fs_token_len=np.median(actual_first_sentence_lens or 0),
        std_fs_token_len=np.std(actual_first_sentence_lens or 0),
        percentiles_fs_token_len=[(p, np.percentile(actual_first_sentence_lens or 0, p))
                             for p in selected_percentiles], 
        mean_ttft_ms=np.mean(ttfts or 0) *
        1000,  # ttfts is empty if streaming is not supported by backend
        std_ttft_ms=np.std(ttfts or 0) * 1000,
        median_ttft_ms=np.median(ttfts or 0) * 1000,
        percentiles_ttft_ms=[(p, np.percentile(ttfts or 0, p) * 1000)
                             for p in selected_percentiles],
        mean_ttfs_ms=np.mean(ttfss or 0) *
        1000,  # ttfss is empty if streaming is not supported by backend
        std_ttfs_ms=np.std(ttfss or 0) * 1000,
        median_ttfs_ms=np.median(ttfss or 0) * 1000,
        percentiles_ttfs_ms=[(p, np.percentile(ttfss or 0, p) * 1000)
                             for p in selected_percentiles],
        mean_tpot_ms=np.mean(tpots or 0) * 1000,
        std_tpot_ms=np.std(tpots or 0) * 1000,
        median_tpot_ms=np.median(tpots or 0) * 1000,
        percentiles_tpot_ms=[(p, np.percentile(tpots or 0, p) * 1000)
                             for p in selected_percentiles],
        mean_itl_ms=np.mean(itls or 0) * 1000,
        std_itl_ms=np.std(itls or 0) * 1000,
        median_itl_ms=np.median(itls or 0) * 1000,
        percentiles_itl_ms=[(p, np.percentile(itls or 0, p) * 1000)
                            for p in selected_percentiles],
        mean_e2el_ms=np.mean(e2els or 0) * 1000,
        std_e2el_ms=np.std(e2els or 0) * 1000,
        median_e2el_ms=np.median(e2els or 0) * 1000,
        percentiles_e2el_ms=[(p, np.percentile(e2els or 0, p) * 1000)
                             for p in selected_percentiles],
    )

    return metrics, actual_output_lens, actual_first_sentence_lens, actual_input_lens


async def benchmark(
    backend: str,
    api_url: str,
    base_url: str,
    model_id: str,
    tokenizer,
    input_requests: List[Tuple[str, int, int, Dict]],
    logprobs: Optional[int],
    best_of: int,
    request_rate: float,
    burstiness: float,
    disable_tqdm: bool,
    profile: bool,
    selected_percentile_metrics: List[str],
    selected_percentiles: List[str],
    ignore_eos: bool,
    gootput_config_dict: Dict[str, float],
    max_concurrency: Optional[int],
    num_warmup_requests: int = 1,
):
    if backend in ASYNC_REQUEST_FUNCS:
        request_func = ASYNC_REQUEST_FUNCS[backend]
    else:
        raise ValueError(f"Unknown backend: {backend}")

    # Test run with first input
    print("Starting initial single prompt test run...")
    test_prompt, test_prompt_len, test_output_len, test_mm_content = (
        input_requests[0])
    if backend != "openai-chat" and test_mm_content is not None:
        # multi-modal benchmark is only available on OpenAI Chat backend.
        raise ValueError(
            "Multi-modal content is only supported on 'openai-chat' backend.")
    test_input = RequestFuncInput(
        model=model_id,
        prompt=test_prompt,
        api_url=api_url,
        prompt_len=test_prompt_len,
        output_len=test_output_len,
        logprobs=logprobs,
        best_of=best_of,
        multi_modal_content=test_mm_content,
        ignore_eos=ignore_eos,
    )
    # test_output = await request_func(request_func_input=test_input)
    # if not test_output.success:
    #     raise ValueError(
    #         "Initial test run failed - Please make sure benchmark arguments "
    #         f"are correctly specified. Error: {test_output.error}")
    # else:
    #     print("Initial test run completed. Starting main benchmark run...")

    if profile:
        print("Starting profiler...")
        profile_input = RequestFuncInput(model=model_id,
                                         prompt=test_prompt,
                                         api_url=base_url + "/start_profile",
                                         prompt_len=test_prompt_len,
                                         output_len=test_output_len,
                                         logprobs=logprobs,
                                         best_of=best_of,
                                         multi_modal_content=test_mm_content,
                                         ignore_eos=ignore_eos)
        profile_output = await request_func(request_func_input=profile_input)
        if profile_output.success:
            print("Profiler started")


    conversation_groups = {}
    for request in input_requests:
        prompt, prompt_len, output_len, metadata = request
        if metadata and "conversation_id" in metadata:
            conv_id = metadata["conversation_id"]
            if conv_id not in conversation_groups:
                conversation_groups[conv_id] = []
            conversation_groups[conv_id].append(request)
        else:
            # For non-conversation requests, treat as separate conversations
            conv_id = f"single_{time.time()}_{random.randint(0, 1000000)}"
            conversation_groups[conv_id] = [request]

    # Sort each conversation's requests by turn index
    for conv_id in conversation_groups:
        conversation_groups[conv_id].sort(
            key=lambda req: req[3].get("turn_index", 0) if req[3] else 0
        )
    # Create a semaphore for max concurrency
    semaphore = asyncio.Semaphore(max_concurrency) if max_concurrency else None

    # Function to process a single conversation
    async def process_conversation(conversation_requests, pbar=None):
        outputs = []
        for prompt, prompt_len, output_len, metadata in conversation_requests:
            request_func_input = RequestFuncInput(
                model=model_id,
                prompt=prompt,
                api_url=api_url,
                prompt_len=prompt_len,
                output_len=output_len,
                logprobs=logprobs,
                best_of=best_of,
                multi_modal_content=metadata.get("multi_modal_content") if metadata else None,
                ignore_eos=ignore_eos,
            )
            
            # Use semaphore if provided
            if semaphore:
                async with semaphore:
                    output = await request_func(request_func_input=request_func_input, 
                                               pbar=pbar)
            else:
                output = await request_func(request_func_input=request_func_input,
                                           pbar=pbar)
            outputs.append(output)
            await asyncio.sleep(0.1)
            return outputs, conversation_requests
    
    if burstiness == 1.0:
        distribution = "Poisson process"
    else:
        distribution = "Gamma distribution"

    print(f"Traffic request rate: {request_rate}")
    print(f"Burstiness factor: {burstiness} ({distribution})")
    print(f"Maximum request concurrency: {max_concurrency}")

    # Group requests by conversation ID
    conversation_ids = list(conversation_groups.keys())
    
    # Generate tasks based on request rate
    if num_warmup_requests > 0:
        warmup_tasks = []
        print(f"Processing {num_warmup_requests} warm up requests")
        warmup_ids = random.choices(conversation_ids, k=num_warmup_requests)
        for conv_id in warmup_ids:
            conversation = conversation_groups[conv_id]
            warmup_tasks.append(asyncio.create_task(process_conversation(conversation)))
        await asyncio.gather(*warmup_tasks)
   
   
   
    print(f"Processing {len(input_requests) } requests")
    benchmark_start_time = time.perf_counter()
    pbar = None if disable_tqdm else tqdm(total=len(input_requests))
    
    tasks = []
    # Create tasks at the specified rate

    if request_rate == float("inf"):
        # Start all conversations at once
        for conv_id in conversation_ids:
            conversation = conversation_groups[conv_id]
            tasks.append(asyncio.create_task(process_conversation(conversation, pbar)))
    else:
        # Start conversations at specified rate
        async def launch_conversations():
            for i, conv_id in enumerate(conversation_ids):
                conversation = conversation_groups[conv_id]
                tasks.append(asyncio.create_task(process_conversation(conversation, pbar)))
                
                if i < len(conversation_ids) - 1:  # Don't wait after the last one
                    # Sample interval based on request rate and burstiness
                    if burstiness == 1.0:
                        # Poisson process - exponential intervals
                        interval = np.random.exponential(1.0 / request_rate)
                    else:
                        # Gamma distribution for controlled burstiness
                        theta = 1.0 / (request_rate * burstiness)
                        interval = np.random.gamma(shape=burstiness, scale=theta)
                    
                    await asyncio.sleep(interval)
        
        await launch_conversations()
    
    # Wait for all tasks to complete
    results = await asyncio.gather(*tasks)

    # Flatten the results for metrics calculation
    all_outputs = []
    all_requests = []
    for outputs, requests in results:
        all_outputs.extend(outputs)
        all_requests.extend(requests)
    
    if pbar:
        pbar.close()
    
    benchmark_duration = time.perf_counter() - benchmark_start_time
    
    # Calculate metrics (ensure the order matches between requests and outputs)
    metrics, actual_output_lens, first_sentence_lens, actual_input_lens = calculate_metrics(
        input_requests=all_requests,
        outputs=all_outputs,
        dur_s=benchmark_duration,
        tokenizer=tokenizer,
        selected_percentile_metrics=selected_percentile_metrics,
        selected_percentiles=selected_percentiles,
        gootput_config_dict=gootput_config_dict,
    )
    
    print("{s:{c}^{n}}".format(s=' Serving Benchmark Result ', n=50, c='='))
    print("{:<40} {:<10}".format("Successful requests:", metrics.completed))
    print("{:<40} {:<10.2f}".format("Benchmark duration (s):",
                                    benchmark_duration))
    print("{:<40} {:<10}".format("Total input tokens:", metrics.total_input))
    print("{:<40} {:<10}".format("Total generated tokens:",
                                 metrics.total_output))
    print("{:<40} {:<10.2f}".format("Request throughput (req/s):",
                                    metrics.request_throughput))
    if gootput_config_dict:
        print("{:<40} {:<10.2f}".format("Request goodput (req/s):",
                                        metrics.request_goodput))
    print("{:<40} {:<10.2f}".format("Output token throughput (tok/s):",
                                    metrics.output_throughput))
    print("{:<40} {:<10.2f}".format("Total Token throughput (tok/s):",
                                    metrics.total_token_throughput))

    result = {
        "duration": benchmark_duration,
        "completed": metrics.completed,
        "total_input_tokens": metrics.total_input,
        "total_output_tokens": metrics.total_output,
        "request_throughput": metrics.request_throughput,
        "request_goodput:":
        metrics.request_goodput if gootput_config_dict else None,
        "output_throughput": metrics.output_throughput,
        "total_token_throughput": metrics.total_token_throughput,
        "input_lens": actual_input_lens,
        "output_lens": actual_output_lens,
        "first_sentence_lens": first_sentence_lens,
        "ttfts": [output.ttft for output in all_outputs],
        "ttfss": [output.ttfs for output in all_outputs],
        # "itls": [output.itl for output in all_outputs],
        "generated_texts": [output.generated_text for output in all_outputs],
        "first_sentence_texts": [output.first_sentence_text for output in all_outputs],
        "e2els": [output.latency for output in all_outputs],
        "errors": [output.error for output in all_outputs],
    }

    def process_one_metric(
        # E.g., "ttft"
        metric_attribute_name: str,
        # E.g., "TTFT"
        metric_name: str,
        # E.g., "Time to First Token"
        metric_header: str,
        # E.g., "ms"
        metric_unit: str,
    ):
        # This function prints and adds statistics of the specified
        # metric.
        if metric_attribute_name not in selected_percentile_metrics:
            return
        print("{s:{c}^{n}}".format(s=metric_header, n=50, c='-'))
        print("{:<40} {:<10.2f}".format(
            f"Mean {metric_name} ({metric_unit}):",
            getattr(metrics, f"mean_{metric_attribute_name}_{metric_unit}")))
        print("{:<40} {:<10.2f}".format(
            f"Median {metric_name} ({metric_unit}):",
            getattr(metrics, f"median_{metric_attribute_name}_{metric_unit}")))
        result[f"mean_{metric_attribute_name}_{metric_unit}"] = getattr(
            metrics, f"mean_{metric_attribute_name}_{metric_unit}")
        result[f"median_{metric_attribute_name}_{metric_unit}"] = getattr(
            metrics, f"median_{metric_attribute_name}_{metric_unit}")
        result[f"std_{metric_attribute_name}_{metric_unit}"] = getattr(
            metrics, f"std_{metric_attribute_name}_{metric_unit}")
        for p, value in getattr(metrics,
                                f"percentiles_{metric_attribute_name}_{metric_unit}"):
            p_word = str(int(p)) if int(p) == p else str(p)
            print("{:<40} {:<10.2f}".format(f"P{p_word} {metric_name} ({metric_unit}):",
                                            value))
            result[f"p{p_word}_{metric_attribute_name}_{metric_unit}"] = value

    process_one_metric("input_token", "Input Tokens", "Stats for Input Tokens Len", "len")
    process_one_metric("output_token", "Output Tokens", "Stats for Output Tokens Len", "len")
    process_one_metric("fs_token", "First Sentence Tokens", "Stats for First Sentence Tokens Len", "len")
    process_one_metric("ttfs", "TTFS", "Time to First Sentence", "ms")
    process_one_metric("ttft", "TTFT", "Time to First Token", "ms")
    process_one_metric("tpot", "TPOT",
                       "Time per Output Token (excl. 1st token)", "ms")
    process_one_metric("itl", "ITL", "Inter-token Latency", "ms")
    process_one_metric("e2el", "E2EL", "End-to-end Latency", "ms")

    print("=" * 50)

    return result


def check_goodput_args(args):
    # Check and parse goodput arguments
    gootput_config_dict = {}
    VALID_NAMES = ["ttft", "ttfs", "tpot", "e2el"]
    if args.goodput:
        gootput_config_dict = parse_goodput(args.goodput)
        for slo_name, slo_val in gootput_config_dict.items():
            if slo_name not in VALID_NAMES:
                raise ValueError(
                    f"Invalid metric name found, {slo_name}: {slo_val}. "
                    "The service level objective name should be one of "
                    f"{str(VALID_NAMES)}. ")
            if slo_val < 0:
                raise ValueError(
                    f"Invalid value found, {slo_name}: {slo_val}. "
                    "The service level objective value should be "
                    "non-negative.")
    return gootput_config_dict


def parse_goodput(slo_pairs):
    gootput_config_dict = {}
    try:
        for slo_pair in slo_pairs:
            slo_name, slo_val = slo_pair.split(":")
            gootput_config_dict[slo_name] = float(slo_val)
    except ValueError as err:
        raise argparse.ArgumentTypeError(
            "Invalid format found for service level objectives. "
            "Specify service level objectives for goodput as \"KEY:VALUE\" "
            "pairs, where the key is a metric name, and the value is a "
            "number in milliseconds.") from err
    return gootput_config_dict





# for now just return a torch dataloader
# TODO: implement the actual anc dataloader
def create_dataloader(*args, **kwargs):
    try:
        import torch
    except ImportError as e:
        raise ImportError(
            "Optional dependency 'torch' is required to use `create_dataloader()`. "
            "Please install it with `pip install torch`."
        ) from e
    return torch.utils.data.DataLoader(*args, **kwargs)



import logging


class SeqSplitConfig:
    """Configuration class for sequence splitting behavior.
    
    Attributes:
        allow_split: Boolean flag to enable/disable sequence splitting
        split_threshold: Minimum sequence length that triggers splitting
        freeze_batch_interval: Number of batches to wait before unfreezing split sequences
    """
    def __init__(
        self,
        allow_split=False,
        split_threshold=None,
        freeze_batch_interval=None
    ):
        self.allow_split = allow_split
        self.split_threshold = split_threshold
        self.freeze_batch_interval = freeze_batch_interval


# AncComposerBuffer uses get_token_length_fn to get the token length of each sample.
# Set get_token_length_fn properly to make sample composing correct.
class AncComposerBuffer:
    """Buffer class that manages token sequences and their composition.
    
    This class handles the storage, splitting, and retrieval of token sequences while
    maintaining proper token length tracking and batch management.
    
    Attributes:
        buffer: List of tuples containing (token_length, item)
        total_token_length: Running sum of all token lengths in buffer
        get_token_length_fn: Function to calculate token length of an item
        split_fn: Function to split sequences that are too long
        delete_idx: Set of indices marked for deletion
        freeze_splits_buffer: Buffer for split sequences that are temporarily frozen
        unfrozen_splits_buffer: Buffer for split sequences ready to be processed
    """
    def __init__(
        self,
        get_token_length_fn=None,
        split_fn=None,
        freeze_batch_interval=None,
        enable_logging=True,
        # get_from_split_chunks_prob=0.6,
    ):
        self.buffer = []
        self.total_token_length = 0
        self.get_token_length_fn = get_token_length_fn
        self.split_fn = split_fn
        self.delete_idx = set()
        self.cur_idx = 0
        # if freeze_batch_interval is None, we do not care the split chunks existing in the same minibatch
        # otherwise, we need to hold the remained splitted chunks for a while to make sure
        # split chunks trained in different mini batches
        self.freeze_batch_interval = freeze_batch_interval
        self.enable_split_freeze = freeze_batch_interval is not None
        self.batch_count = 0
        self.freeze_splits_buffer = []
        self.unfrozen_splits_buffer = []
        self.enable_logging = enable_logging
        # self.random_state = None
        # self.get_from_split_chunks_prob = get_from_split_chunks_prob

    def put(self, item):
        """Add an item to the buffer and return total token length.
        
        Args:
            item: The sequence item to add to the buffer
        Returns:
            int: Updated total token length in buffer
        """
        if item is not None:
            token_length = self._get_token_length(item)
            self.buffer.append((token_length, item))
            self.total_token_length += token_length
        return self.total_token_length

    def get(self, target_length, first_call=False):
        """Retrieve first sample that fits within target length.
        
        Args:
            target_length: Maximum allowed token length
            first_call: Whether this is the first retrieval attempt for current batch
        Returns:
            tuple: (token_length, item) or None if no suitable item found
        """
        # try to get from split chunks first
        if first_call and self.split_fn is not None and self.enable_split_freeze:
            # when first_call is True, the target_length should be equal to context window size
            split_chunk = self.try_to_get_from_split_chunks(target_length)
            if split_chunk is not None:
                return split_chunk
        for i in range(len(self.buffer)):
            idx = (i + self.cur_idx) % len(self.buffer)
            if idx in self.delete_idx:
                continue
            if self.buffer[idx][0] <= target_length:
                self.delete_idx.add(idx)
                self.total_token_length -= self.buffer[idx][0]
                self.cur_idx = (idx + 1) % len(self.buffer)
                return self.buffer[idx]
        return None

    def _get_token_length(self, item):
        """Helper function to get token length of an item.
        
        Args:
            item: The item to get token length of
        Returns:
            int: Token length of the item, default to 1 if get_token_length_fn is not provided
        """
        if self.get_token_length_fn is not None:
            return self.get_token_length_fn(item)
        else:
            return 1

    # try to get target_length from unfrozen split chunks
    def try_to_get_from_split_chunks(self, target_length):
        """Try to get a split chunk that fits within target length.
        
        Args:
            target_length: The target length to get
        Returns:
            Tuple: (target_length, split_chunk_of_item) or None if no suitable split chunk found
        """
        if len(self.unfrozen_splits_buffer) == 0:
            return
        # if self.random_state is None:
        #     self.random_state = random.Random(self.batch_count)
        # if self.random_state.random() > self.get_from_split_chunks_prob:
        #     return
        split_chunk = self.unfrozen_splits_buffer.pop(0)
        if split_chunk[0] <= target_length:
            return split_chunk
        else:
            if self.enable_logging:
                print(
                    f"[Composer] chunked item with size {split_chunk[0]} "
                    f"will be split one more time to get {target_length}"
                )
            # the split is still too large, needs to further split it
            to_return, remain = self.split_fn(split_chunk[1], target_length)
            remain_length = self._get_token_length(remain)
            assert self._get_token_length(to_return) == target_length
            self.freeze_splits_buffer.append((remain_length, remain, self.batch_count))
            return (target_length, to_return)
    
    def get_with_split(self, target_length, ignore_split_freezing=False):
        """Split a long sequence to fit target length requirement.
        
        Finds the longest sequence in buffer and splits it to create a sequence
        of exactly target_length tokens.
        
        Args:
            target_length: Desired length of the returned sequence
            ignore_split_freezing: Whether to bypass freeze mechanism for splits
        Returns:
            tuple: (target_length, split_chunk_of_item) or None if splitting not possible
        """
        if self.total_token_length == 0:
            return None
        if self.split_fn is None:
            return None
        max_length_idx = -1
        max_length = target_length
        for i, item in enumerate(self.buffer):
            if i in self.delete_idx:
                continue
            if item[0] > max_length:
                max_length = item[0]
                max_length_idx = i
        if max_length_idx == -1:
            return None
        assert max_length > target_length, (
            f"Invalid buffer {[i[0] for i in self.buffer]} with target_length {target_length}"
        )
        assert self.buffer[max_length_idx][0] == max_length
        to_return, remain = self.split_fn(self.buffer[max_length_idx][1], target_length)
        remain_length = self._get_token_length(remain)
        assert self._get_token_length(to_return) == target_length
        
        if not self.enable_split_freeze or ignore_split_freezing:
            self.buffer[max_length_idx] = (remain_length, remain)
            self.total_token_length = (
                self.total_token_length - max_length + remain_length
            )
            self.cur_idx = max_length_idx
        else:
            self.freeze_splits_buffer.append((remain_length, remain, self.batch_count))
            self.total_token_length -= max_length
            self.delete_idx.add(max_length_idx)
        return (target_length, to_return)
    
    def get_all(self):
        """Retrieve all items from the buffer.
        
        Returns:
            list: List of all items in the buffer
        """
        res = self.buffer
        self.buffer = []
        self.total_token_length = 0
        self.delete_idx = set()
        self.cur_ids = 0
        return res
    
    def unfrozen_splits(self):
        """Unfreeze split chunks.
        
        Adds split chunks from freeze_splits_buffer to unfrozen_splits_buffer.
        """
        to_freeze = [(i[0], i[1]) for i in self.freeze_splits_buffer if self.batch_count - i[2] > self.freeze_batch_interval]
        self.unfrozen_splits_buffer += to_freeze
        self.freeze_splits_buffer = [i for i in self.freeze_splits_buffer if self.batch_count - i[2] <= self.freeze_batch_interval]
        if self.enable_logging:
            print(
                f"[Composer] item count in unfrozen_splits_buffer: "
                f"{len(self.unfrozen_splits_buffer)}"
            )
            print(
                f"[Composer] token count in unfrozen_splits_buffer: "
                f"{sum(i[0] for i in self.unfrozen_splits_buffer)}"
            )

    def flush(self):
        """Flush the buffer.
        
        Removes deleted items and updates the buffer.
        It is usually called when finish getting one micro batch of data.
        """
        buffer = [self.buffer[i] for i in range(len(self.buffer)) if i not in self.delete_idx]
        self.buffer = buffer
        self.delete_idx = set()
        #print(f"=====self.total_token_length: {self.total_token_length}, {[i[0] for i in self.unfrozen_splits_buffer]}, {[(i[0], i[2]) for i in self.freeze_splits_buffer]}, {self.batch_count}, {self.freeze_batch_interval}")
        assert sum([i[0] for i in self.buffer]) == self.total_token_length
        if self.enable_split_freeze:
            self.batch_count += 1
            # if self.batch_count == self.freeze_batch_interval:
            #     self.batch_count = 0
            self.unfrozen_splits()
        #print(f"=====self.total_token_length: {self.total_token_length}, {[i[0] for i in self.unfrozen_splits_buffer]}, {[(i[0], i[2]) for i in self.freeze_splits_buffer]}, {self.batch_count}, {self.freeze_batch_interval}")

    def combine_buffers(self):
        """Combine buffers.
        
        Adds split chunks from unfrozen_splits_buffer and freeze_splits_buffer to the main buffer.
        It is usually called when running to the end of the dataset
        """
        self.buffer += self.unfrozen_splits_buffer + [(i[0], i[1]) for i in self.freeze_splits_buffer]
        self.unfrozen_splits_buffer = []
        self.freeze_splits_buffer = []
        self.total_token_length = sum([i[0] for i in self.buffer])

    def _get_ckpt(self):
        states = {}
        states['buffer'] = self.buffer
        states['total_token_length'] = self.total_token_length
        states['unfrozen_splits_buffer'] = self.unfrozen_splits_buffer
        states['freeze_splits_buffer'] = self.freeze_splits_buffer
        states['delete_idx'] = self.delete_idx
        states['cur_idx'] = self.cur_idx
        states['batch_count'] = self.batch_count
        # states['random_state'] = self.random_state
        return states

    def _set_ckpt(self, state):
        self.buffer = state['buffer']
        self.total_token_length = state['total_token_length']
        self.unfrozen_splits_buffer = state['unfrozen_splits_buffer']
        self.freeze_splits_buffer = state['freeze_splits_buffer']
        self.delete_idx = state['delete_idx']
        self.cur_idx = state['cur_idx']
        self.batch_count = state['batch_count']
        # self.random_state = state['random_state']

class AncComposer:
    """Main composer class that handles sequence composition and batch creation.
    
    This class coordinates the buffer management, sequence splitting, and batch
    formation according to specified constraints.
    
    Attributes:
        max_seq_len: Maximum allowed sequence length
        allow_split: Whether to allow splitting of long sequences
        ratio: Controls buffer size relative to max_seq_len
        split_threshold: Minimum length that triggers sequence splitting
    """
    def __init__(
        self,
        max_seq_len,
        get_token_length_fn=None,
        ratio=2.0,
        seq_split_config=None,
        split_fn=None,
        enable_logging=True,
    ):
        self.max_seq_len = max_seq_len
        self.allow_split = seq_split_config.allow_split if seq_split_config is not None else False
        freeze_batch_interval = seq_split_config.freeze_batch_interval if seq_split_config is not None else None
        if self.allow_split:
            assert split_fn is not None, "Need to provide split_fn to enable allow_split"
        self.split_fn = split_fn if self.allow_split else None
        self.buffer = AncComposerBuffer(
            get_token_length_fn,
            self.split_fn,
            freeze_batch_interval,
            enable_logging,
        )
        # ratio here controls how many samples would be collected in the buffer
        # it might be difficult to compose enough samples to reach the max_seq_len when buffer size is small
        self.ratio = ratio
        self.split_threshold = seq_split_config.split_threshold if seq_split_config is not None else max_seq_len
        if self.split_threshold is None:
            self.split_threshold = max_seq_len
        self.enable_logging = enable_logging
        self.remain = []

    # function to return one micro batch
    def get_items(self, is_last_call=False, ignore_split_freezing=False):
        """Retrieve one micro batch (max_seq_len of tokens) from the buffer.
        
        Args:
            is_last_call: Whether this is the final call (triggers buffer cleanup)
            ignore_split_freezing: Whether to bypass freeze mechanism for splits
        Returns:
            list: List of items in the micro batch, the sum of the token length of the items should be equal to or less than max_seq_len
        """
        items = []
        if is_last_call:
            items = self.buffer.get_all()
            items = [i[1] for i in items if i is not None]
            return items
        cur_seq_length = self.max_seq_len
        first_call = True
        while True:
            item = self.buffer.get(cur_seq_length, first_call)
            first_call = False
            if item is None:
                break
            assert len(item) == 2
            items.append(item[1])
            cur_seq_length = cur_seq_length - item[0]
            if cur_seq_length == 0:
                break
        if cur_seq_length >= self.split_threshold and self.allow_split:
            item = self.buffer.get_with_split(cur_seq_length, ignore_split_freezing)
            if item is not None:
                assert item[0] == cur_seq_length
                items.append(item[1])
        self.buffer.flush()
        return items

    def apply(self, sample, is_last_call=False):
        """Process a single sample and generate batches if buffer is full enough.
        
        Args:
            sample: Input item to process
            is_last_call: Whether this is the final call (triggers buffer cleanup)
        Returns:
            list: List of batched items or None if no batches were created
        """
        total_token_length = self.buffer.put(sample)
        res = []
        while total_token_length >= self.max_seq_len * self.ratio:
            items = self.get_items()
            res.append(items)
            total_token_length = self.buffer.put(None)
        if is_last_call:
            self.buffer.combine_buffers()
            total_token_length = self.buffer.put(None)
            while total_token_length >= self.max_seq_len:
                items = self.get_items(ignore_split_freezing=True)
                res.append(items)
                total_token_length = self.buffer.put(None)
            if total_token_length > 0:
                items = self.get_items(is_last_call=True)
                res.append(items)
        return res if len(res) > 0 else None
    
    def __call__(self, samples, is_last_call=False):
        """Entrypoint of the composer.
        
        Args:
            samples: List of samples to process
            is_last_call: Whether this is the final call (triggers buffer cleanup)
        Returns:
            list: List of batched items or None if no batches were created
        """
        if self.remain:
            yield from self.remain
            self.remain = []
        for sample in samples:
            if sample is None:
                continue
            res = self.apply(sample)
            if res is None:
                continue
            for i, item in enumerate(res):
                self.remain = res[i+1:]
                yield item
            # yield from res
        assert len(self.remain) == 0
        if is_last_call:
            res = self.apply(None, True)
            if res is not None:
                yield from res

    def _get_ckpt(self):
        states = {}
        states['buffer'] = self.buffer._get_ckpt()
        states['remain'] = self.remain
        return states
    
    def _set_ckpt(self, state):
        self.buffer._set_ckpt(state['buffer'])
        if 'remain' in state:
            self.remain = state['remain']


if __name__ == "__main__":
    import random
    from functools import partial
    def create_fake_data(min_token_length, max_seq_length, max_token_length):
        i  = 0
        while True:
            if random.random() <= 0.2:
                data = random.randint(max_seq_length, max_token_length)
            else:
                data = random.randint(min_token_length, max_seq_length)
            yield {'doc_id': i, "tokens": data}
            i += 1
    
    def fake_split_fn(item, split_length, prefix_len):
        to_return = {}
        remain = {}
        to_return['doc_id'] = item['doc_id']
        remain['doc_id'] = item['doc_id']
        to_return['tokens'] = split_length
        assert split_length > prefix_len
        remain['tokens'] = item['tokens'] + prefix_len - split_length
        return to_return, remain
    
    def fake_get_token_length_fn(item):
        return item['tokens']
    
    max_seq_length = 8192  # 8K
    freeze_batch_interval = 64
    split_threshold = 128
    prefix_len = 32
    min_token_length = prefix_len + 32
    max_token_length = 2 ** 15  # 32K
    batch_limit = 100

    
    split_config = SeqSplitConfig(
        allow_split=True,
        split_threshold=split_threshold,
        freeze_batch_interval=freeze_batch_interval,
    )
    
    composer = AncComposer(
        max_seq_length,
        get_token_length_fn=fake_get_token_length_fn,
        seq_split_config=split_config,
        split_fn=partial(fake_split_fn, prefix_len=prefix_len),
    )

    data_generator = create_fake_data(min_token_length, max_seq_length, max_token_length)
    micro_batch_count = 0
    batch_doc_id = []
    token_lengths = []
    batch_count = 0
    for data in data_generator:
        res = composer([data])
        for micro_batch in res:
            micro_batch_count += 1
            doc_ids = [i['doc_id'] for i in micro_batch]
            token_length = sum([i['tokens'] for i in micro_batch])
            token_lengths.append(token_length)
            assert max_seq_length - split_threshold < token_length <= max_seq_length
            batch_doc_id += doc_ids
            if micro_batch_count == freeze_batch_interval:
                assert len(batch_doc_id) == len(set(batch_doc_id))
                print(token_lengths, batch_doc_id)
                micro_batch_count = 0
                batch_doc_id = []
                batch_count += 1
                token_lengths = []
        if batch_count >= batch_limit:
            break



import click
import logging
import sys
from anc.version import __version__

from anc.cli import deployment
from anc.cli import dataset
from anc.cli import quantization
from anc.cli import evaluation

from anc.cli.load_testing import loadtest
from anc.cli.util import click_group

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    handlers=[
        logging.StreamHandler(sys.stdout)
    ]
)
logger = logging.getLogger("anc")


@click.version_option(__version__, "-v", "--version")
@click.option('--log-level', 
              type=click.Choice(['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], 
                               case_sensitive=False),
              default='INFO',
              help='Set the logging level')
@click_group(context_settings=CONTEXT_SETTINGS)
def main(log_level):
    # Configure logging with specified level
    logging.basicConfig(
        level=getattr(logging, log_level.upper()),
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        handlers=[logging.StreamHandler(sys.stdout)]
    )
    pass


#deployment.add_command(main)
dataset.add_command(main)
quantization.add_command(main)
evaluation.add_command(main)
main.add_command(loadtest)

if __name__ == "__main__":
    main()



import json
import numpy as np
import os
import traceback
def read_benchmark_results(file_path):
    with open(file_path, "r") as f:
        return json.load(f)

def analyze_benchmark_results(file_path, base_result_dir, first_sentence_len=50):
    """
    Analyze benchmark results by creating a pandas DataFrame and generating both static PNG plots and interactive HTML dashboard.
    
    Args:
        file_path: Path to benchmark results file or directory
        base_result_dir: Directory to save analysis results
    """
    try:
        import pandas as pd
        import matplotlib.pyplot as plt
        import numpy as np
    except ImportError as e:
        print("Missing optional dependencies for analysis: pandas, matplotlib, numpy")
        print("Install them with: pip install pandas matplotlib numpy")
        return
    
    try:
        # Check if file_path is a directory or a single file
        all_results = []
        if os.path.isdir(file_path):
            # Process all JSON files in the directory
            print(f"Processing all JSON files in the directory: {file_path}")
            json_files = [f for f in os.listdir(file_path) if f.endswith('.json')]
            for json_file in json_files:
                full_path = os.path.join(file_path, json_file)
                file_results = read_benchmark_results(full_path)
                # Add filename as a column (without .json extension)
                file_name = os.path.basename(json_file).replace('.json', '')
                for result in file_results:
                    result['file_name'] = file_name
                all_results.extend(file_results)
        else:
            # Process a single file
            file_results = read_benchmark_results(file_path)
            file_name = os.path.basename(file_path).replace('.json', '')
            for result in file_results:
                result['file_name'] = file_name
            all_results.extend(file_results)
            
        # Convert list of dictionaries to DataFrame
        df = pd.DataFrame(all_results)
        # extra_info_columns = [col for col in df.columns if not isinstance(df[col].iloc[0], list) and df[col].nunique() == 1]
        # extra_info_columns = []
        # df_extra_info = df[extra_info_columns]
        is_random_data_set = "random_input_len" in df.columns

        # Recalculate request_throughput and output_throughput based on tensor_parallel_size
        if 'tensor_parallel_size' in df.columns and 'request_throughput' in df.columns:
            # Convert tensor_parallel_size to numeric type to avoid TypeError
            df['tensor_parallel_size'] = pd.to_numeric(df['tensor_parallel_size'])
            df['request_throughput'] = df['request_throughput'] * 8 / df['tensor_parallel_size']
        
        if 'tensor_parallel_size' in df.columns and 'output_throughput' in df.columns:
            # Ensure tensor_parallel_size is numeric
            if not pd.api.types.is_numeric_dtype(df['tensor_parallel_size']):
                df['tensor_parallel_size'] = pd.to_numeric(df['tensor_parallel_size'])
            df['output_throughput'] = df['output_throughput'] * 8 / df['tensor_parallel_size']

        grid_search_columns = ["enable_chunked_prefill",
                               "enable_prefix_caching",
                               "use-v2-block-manager",
                               "multi-step-stream-outputs",
                               "tensor_parallel_size",
                               "speculative-draft-tensor-parallel-size",
                               "num-speculative-tokens"]
        if is_random_data_set:
            grid_search_columns.append("random_prefix_len")
        
        select_columns = ["request_throughput",
                          "output_throughput", 
                          "mean_output_token_len",
                          "p90_output_token_len",
                          "mean_input_token_len",
                          "p90_input_token_len",
                          "max_concurrency", 
                          "mean_ttfs_ms",
                          "std_ttfs_ms",
                          "p90_ttfs_ms",
                          "mean_tpot_ms",
                          "std_tpot_ms",
                          "p90_tpot_ms",
                          "mean_ttft_ms",
                          "std_ttft_ms",
                          "p90_ttft_ms",
                          "mean_e2el_ms",
                          "std_e2el_ms",
                          "p90_e2el_ms",
                          "mean_fs_token_len",
                          "std_fs_token_len",
                          "p90_fs_token_len",
         ]

        if 'file_name' in df.columns:
            select_columns = ["file_name"] + select_columns

        if is_random_data_set:
            recalculate_ttfs(df, first_sentence_len=first_sentence_len)
        
        # Filter out columns that have only a single unique value
        all_selected_columns = grid_search_columns + select_columns
        df = df[df['request_throughput'] != 0]  
        df = df[df['tensor_parallel_size'] == 4]
        # df.sort_values(by='request_throughput', ascending=False, inplace=True)


        filtered_columns = []
        for col in all_selected_columns:
            if col in df.columns:
                if df[col].nunique() > 1 or len(df) == 1:
                    filtered_columns.append(col)
        df = df[filtered_columns]
        
        # Keep actual mean_output_token_len values instead of hardcoding to 100
        if 'mean_output_token_len' in df.columns:
            df['mean_output_token_len'] = pd.to_numeric(df['mean_output_token_len'], errors='coerce')
            df['mean_output_token_len'] = df['mean_output_token_len'].round(0).astype(int)
        
        if 'mean_input_token_len' in df.columns:
            df['mean_input_token_len'] = pd.to_numeric(df['mean_input_token_len'], errors='coerce')
            df['mean_input_token_len'] = df['mean_input_token_len'].round(0).astype(int)
        
        df.to_csv(os.path.join(base_result_dir, "benchmark_results.csv"), index=False)
        print(f"Data frame of Benchmark results saved to {os.path.join(base_result_dir, 'benchmark_results.csv')}")
        
        # Rename columns
        renamed_grid_search_columns = {
            "enable_chunked_prefill": "chunkfill",
            "enable_prefix_caching": "precache",
            "tensor_parallel_size": "TP",
            "file_name": "file",
            "speculative-draft-tensor-parallel-size": "sp_tp",
            "num-speculative-tokens": "k",
        }
        
        if is_random_data_set:
            renamed_grid_search_columns["random_prefix_len"] = "prefix_len"
            renamed_grid_search_columns["mean_input_token_len"] = "input_len"
            renamed_grid_search_columns["mean_output_token_len"] = "output_len"
            
        df.rename(columns=renamed_grid_search_columns, inplace=True)
        
        # Create configuration ID for grouping
        config_columns = [col for col in renamed_grid_search_columns.values() if col in df.columns]
        df['config_id'] = df.apply(lambda row: '_'.join([f"{col}_{row[col]}" for col in config_columns]), axis=1)
        
        # Get unique configuration IDs
        config_ids = df['config_id'].unique()
        
        # Create groups based on configuration
        groups = []
        for config_id in config_ids:
            group_df = df[df['config_id'] == config_id]
            if not group_df.empty:
                groups.append((config_id, group_df))
        
        # Use different colors for different configurations
        colors = plt.cm.viridis(np.linspace(0, 1, len(groups)))
        
        # Define metrics to plot
        mean_metrics = [
            ('mean_ttft_ms', 'Mean TTFT (ms)'),
            ('mean_tpot_ms', 'Mean TPOT (ms)'),
            ('mean_ttfs_ms', 'Mean TTFS (ms)'),
            ('mean_e2el_ms', 'Mean E2EL (ms)')
        ]
        
        p90_metrics = [
            ('p90_ttft_ms', 'P90 TTFT (ms)'),
            ('p90_tpot_ms', 'P90 TPOT (ms)'),
            ('p90_ttfs_ms', 'P90 TTFS (ms)'),
            ('p90_e2el_ms', 'P90 E2EL (ms)')
        ]
        
        # Function to create matplotlib plots with given x-axis
        def create_performance_plots(y_axis, y_label, filename, swap_axes=False):
            # default x-axis is max_concurrency
            fig, axes = plt.subplots(2, 4, figsize=(20, 10))
            
            # Flatten axes for easier iteration
            axes = axes.flatten()
            
            # Helper function to create a plot for a metric
            def plot_metric(ax, metric, metric_label, y_axis_name, groups_data, swap_axes=swap_axes):
                # Define different marker styles for different configurations
                markers = ['o', 's', '^', 'D', 'v', '<', '>', 'p', '*', 'h', 'H', '+', 'x', '|']
                
                for j, (config_id, group) in enumerate(groups_data):
                    # Use different marker styles based on configuration
                    marker_idx = j % len(markers)
                    marker_style = markers[marker_idx]
                    
                    if swap_axes:
                        ax.scatter(group[y_axis], group[metric],
                                  label=config_id, 
                                  color=colors[j], marker=marker_style, s=70)
                        ax.plot(group[y_axis], group[metric],
                               color=colors[j], linestyle='-')
                        ax.set_xlabel(y_label)
                        ax.set_ylabel(metric_label)
                        ax.set_title(f'{y_label} vs {metric_label}')
                    else:
                        ax.scatter(group[metric], group[y_axis],
                                  label=config_id, 
                                  color=colors[j], marker=marker_style, s=70)
                        ax.plot(group[metric], group[y_axis],
                               color=colors[j], linestyle='-')
                        ax.set_ylabel(y_label)
                        ax.set_xlabel(metric_label)
                        ax.set_title(f'{metric_label} vs {y_label}')
                ax.grid(True)
                
                # Only add legend to the first graph (index 0)
                if ax == axes[0]:
                    ax.legend(fontsize='x-small', loc='upper right', frameon=True)
            
            # Plot mean metrics (first row)
            for i, (metric, ylabel) in enumerate(mean_metrics):
                plot_metric(axes[i], metric, ylabel, y_label, groups, swap_axes=swap_axes)
            
            # Plot p90 metrics (second row)
            for i, (metric, ylabel) in enumerate(p90_metrics):
                plot_metric(axes[i+4], metric, ylabel, y_label, groups, swap_axes=swap_axes)
            
            plt.tight_layout()
            plt.savefig(os.path.join(base_result_dir, filename))
            plt.close()
            
        # Create matplotlib plots
        create_performance_plots('max_concurrency', 'Concurrency', 'concurrency_analysis.png', swap_axes=True)
        create_performance_plots('request_throughput', 'Request Throughput', 'throughput_analysis.png')
        create_interactive_dashboard(df, base_result_dir)
        
    except ImportError as e:
        print(f"Could not analyze results: {e}")
        print("To enable analysis, install required packages: pip install pandas matplotlib plotly")
    except Exception as e:
        print(f"Error analyzing benchmark results: {e}")
        print(traceback.format_exc())
        print(traceback.format_exc())

def create_interactive_dashboard(df, base_result_dir):
    """
    Create an interactive HTML dashboard with Plotly and dropdown filters.
    """
    try:
        import plotly.graph_objects as go
        import plotly.express as px
        from plotly.subplots import make_subplots
        import plotly.offline as pyo
    except ImportError:
        print("Plotly not available, skipping interactive dashboard creation")
        return
        
    # Create plots directory
    plot_dir = os.path.join(base_result_dir)
    
    # Convert DataFrame to JSON for JavaScript
    data_json = df.to_json(orient='records')
    
    # Get configuration columns (exclude metric columns)
    metric_columns = ['request_throughput', 'output_throughput', 'mean_ttft_ms', 'mean_tpot_ms', 
                      'mean_e2el_ms', 'mean_ttfs_ms', 'std_ttft_ms', 'std_tpot_ms', 'std_e2el_ms', 
                      'std_ttfs_ms', 'p90_ttft_ms', 'p90_tpot_ms', 'p90_e2el_ms', 'p90_ttfs_ms', 
                      'mean_fs_token_len', 'std_fs_token_len', 'p90_fs_token_len', 'max_concurrency',
                      'p90_input_token_len', 'config_id']
    config_columns = [col for col in df.columns if col not in metric_columns]
    
    # Get unique values for filters
    filter_options = {}
    for col in config_columns:
        unique_vals = sorted(df[col].unique())
        filter_options[col] = unique_vals
    
    # Define metrics to plot - Updated to include TTFS and remove request_throughput vs request_throughput
    metrics = [
        ['mean_ttft_ms', 'Mean TTFT (ms)'],
        ['mean_tpot_ms', 'Mean TPOT (ms)'],
        ['mean_ttfs_ms', 'Mean TTFS (ms)'],
        ['mean_e2el_ms', 'Mean E2E Latency (ms)']
    ]
    
    p90_metrics = [
        ['p90_ttft_ms', 'P90 TTFT (ms)'],
        ['p90_tpot_ms', 'P90 TPOT (ms)'],
        ['p90_ttfs_ms', 'P90 TTFS (ms)'],
        ['p90_e2el_ms', 'P90 E2E Latency (ms)']
    ]

    # Generate filter controls HTML
    filter_controls_html = ""
    for col, values in filter_options.items():
        filter_controls_html += f"""
            <div class="filter-group">
                <label for="{col}_filter">{col.replace('_', ' ').title()}:</label>
                <select id="{col}_filter" onchange="updatePlots()">
                    <option value="all">All</option>
                    {''.join([f'<option value="{val}">{val}</option>' for val in values])}
                </select>
            </div>
"""

    # Generate unique values summary HTML
    unique_values_html = ""
    for col, values in filter_options.items():
        values_str = ", ".join([str(val) for val in values])
        unique_values_html += f"""
            <tr>
                <td class="column-name">{col.replace('_', ' ').title()}</td>
                <td class="unique-count">{len(values)}</td>
                <td class="values-list">{values_str}</td>
            </tr>
"""

    html_content = f"""
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Interactive Benchmark Dashboard</title>
    <script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
    <style>
        body {{
            font-family: Arial, sans-serif;
            margin: 0;
            padding: 20px;
            background-color: #f5f5f5;
        }}
        .container {{
            max-width: 1400px;
            margin: 0 auto;
            background-color: white;
            padding: 20px;
            border-radius: 10px;
            box-shadow: 0 2px 10px rgba(0,0,0,0.1);
        }}
        .title {{
            text-align: center;
            color: #333;
            margin-bottom: 30px;
        }}
        .controls {{
            display: flex;
            flex-wrap: wrap;
            gap: 15px;
            margin-bottom: 30px;
            padding: 20px;
            background-color: #f8f9fa;
            border-radius: 8px;
        }}
        .filter-group {{
            display: flex;
            flex-direction: column;
            gap: 5px;
        }}
        .filter-group label {{
            font-weight: bold;
            color: #555;
        }}
        .filter-group select {{
            padding: 8px 12px;
            border: 1px solid #ddd;
            border-radius: 4px;
            background-color: white;
        }}
        .unique-values-section {{
            margin-bottom: 30px;
            padding: 20px;
            background-color: #fff;
            border-radius: 8px;
            box-shadow: 0 1px 5px rgba(0,0,0,0.1);
        }}
        .unique-values-section h3 {{
            margin-top: 0;
            margin-bottom: 20px;
            color: #333;
            text-align: center;
        }}
        .unique-values-table {{
            width: 100%;
            border-collapse: collapse;
            margin-top: 15px;
        }}
        .unique-values-table th {{
            background-color: #f8f9fa;
            padding: 12px;
            text-align: left;
            border-bottom: 2px solid #dee2e6;
            font-weight: bold;
            color: #495057;
        }}
        .unique-values-table td {{
            padding: 10px 12px;
            border-bottom: 1px solid #dee2e6;
            vertical-align: top;
        }}
        .unique-values-table tr:hover {{
            background-color: #f8f9fa;
        }}
        .column-name {{
            font-weight: bold;
            color: #495057;
            min-width: 150px;
        }}
        .unique-count {{
            text-align: center;
            font-weight: bold;
            color: #007bff;
            min-width: 80px;
        }}
        .values-list {{
            word-wrap: break-word;
            max-width: 600px;
        }}
        .axis-controls {{
            margin-bottom: 20px;
            padding: 15px;
            background-color: #e9ecef;
            border-radius: 8px;
        }}
        .axis-controls label {{
            font-weight: bold;
            margin-right: 10px;
        }}
        .axis-controls select {{
            padding: 8px 12px;
            border: 1px solid #ddd;
            border-radius: 4px;
            background-color: white;
        }}
        .plots-container {{
            display: flex;
            flex-direction: column;
            gap: 30px;
        }}
        .plot-section {{
            background-color: white;
            border-radius: 8px;
            padding: 20px;
            box-shadow: 0 1px 5px rgba(0,0,0,0.1);
        }}
        .plot-section h3 {{
            margin-top: 0;
            margin-bottom: 20px;
            color: #333;
            text-align: center;
        }}
        .plot-grid {{
            display: grid;
            grid-template-columns: repeat(2, 1fr);
            gap: 20px;
        }}
        .plot-item {{
            min-height: 400px;
            border: 1px solid #eee;
            border-radius: 8px;
        }}
    </style>
</head>
<body>
    <div class="container">
        <h1 class="title">Interactive Benchmark Results Dashboard</h1>
        
        <div class="unique-values-section">
            <h3>Dataset Configuration Summary</h3>
            <table class="unique-values-table">
                <thead>
                    <tr>
                        <th>Configuration Parameter</th>
                        <th>Unique Values Count</th>
                        <th>Available Values</th>
                    </tr>
                </thead>
                <tbody>
                    {unique_values_html}
                </tbody>
            </table>
        </div>
        
        <div class="controls">
            {filter_controls_html}
        </div>
        
        <div class="axis-controls">
            <label for="x_axis_select">Chart Type:</label>
            <select id="x_axis_select" onchange="updatePlots()">
                <option value="max_concurrency">Metrics vs Concurrency</option>
                <option value="request_throughput">Throughput vs Metrics</option>
            </select>
        </div>
        
        <div class="plots-container">
            <div class="plot-section">
                <h3>Mean Metrics vs X-Axis</h3>
                <div class="plot-grid">
                    <div id="mean_plot_0" class="plot-item"></div>
                    <div id="mean_plot_1" class="plot-item"></div>
                    <div id="mean_plot_2" class="plot-item"></div>
                    <div id="mean_plot_3" class="plot-item"></div>
                </div>
            </div>
            
            <div class="plot-section">
                <h3>P90 Metrics vs X-Axis</h3>
                <div class="plot-grid">
                    <div id="p90_plot_0" class="plot-item"></div>
                    <div id="p90_plot_1" class="plot-item"></div>
                    <div id="p90_plot_2" class="plot-item"></div>
                    <div id="p90_plot_3" class="plot-item"></div>
                </div>
            </div>
        </div>
    </div>

    <script>
        // Data from Python
        const rawData = {data_json};
        
        // Metrics definitions
        const metrics = {json.dumps(metrics)};
        const p90_metrics = {json.dumps(p90_metrics)};
        const configColumns = {json.dumps(config_columns)};
        
        // Global variable to track legend visibility state
        let legendVisibility = {{}};
        
        // Dynamically detect configuration columns (excluding metric columns)
        const metricColumns = ['request_throughput', 'output_throughput', 'mean_ttft_ms', 'mean_tpot_ms', 
                              'mean_e2el_ms', 'mean_ttfs_ms', 'std_ttft_ms', 'std_tpot_ms', 'std_e2el_ms', 
                              'std_ttfs_ms', 'p90_ttft_ms', 'p90_tpot_ms', 'p90_e2el_ms', 'p90_ttfs_ms', 
                              'mean_fs_token_len', 'std_fs_token_len', 'p90_fs_token_len', 'max_concurrency',
                              'p90_input_token_len', 'config_id'];
        
        function getConfigColumns() {{
            if (rawData.length === 0) return [];
            const allColumns = Object.keys(rawData[0]);
            return allColumns.filter(col => !metricColumns.includes(col));
        }}
        
        function filterData() {{
            let filteredData = rawData.slice();
            
            // Apply filters dynamically based on available config columns
            configColumns.forEach(col => {{
                const filterElement = document.getElementById(col + '_filter');
                if (filterElement) {{
                    const filterValue = filterElement.value;
                    if (filterValue !== 'all') {{
                        filteredData = filteredData.filter(row => row[col].toString() === filterValue);
                    }}
                }}
            }});
            
            return filteredData;
        }}
        
        function createTrace(data, xAxis, yMetric) {{
            const groups = {{}};
            const configColumns = getConfigColumns();
            
            data.forEach(row => {{
                // Create a flexible key based on available config columns
                const keyParts = configColumns.map(col => {{
                    if (row[col] !== undefined && row[col] !== null) {{
                        return `${{col}}:${{row[col]}}`;
                    }}
                    return null;
                }}).filter(part => part !== null);
                
                const key = keyParts.join(' | ');
                
                if (!groups[key]) {{
                    // Check if this trace should be visible based on global state
                    const visible = legendVisibility[key] !== false;
                    
                    groups[key] = {{
                        x: [],
                        y: [],
                        text: [],
                        name: key,
                        type: 'scatter',
                        mode: 'markers+lines',
                        visible: visible,
                        hovertemplate: '<b>%{{text}}</b><br>' +
                                     'X: %{{x}}<br>' +
                                     'Y: %{{y}}<br>' +
                                     '<extra></extra>',
                        hoverlabel: {{
                            bgcolor: 'white',
                            bordercolor: 'black',
                            font: {{size: 12}},
                            namelength: -1
                        }}
                    }};
                }}
                groups[key].x.push(row[xAxis]);
                groups[key].y.push(row[yMetric]);
                groups[key].text.push(key); // Add the full key as hover text for each point
            }});
            
            return Object.values(groups);
        }}
        
        function setupLegendClickHandler(plotId) {{
            document.getElementById(plotId).on('plotly_legendclick', function(data) {{
                const traceName = data.data[data.curveNumber].name;
                
                // Toggle visibility state
                legendVisibility[traceName] = !legendVisibility[traceName];
                
                // Update all plots with new visibility
                updateAllPlotsVisibility();
                
                // Prevent default legend behavior
                return false;
            }});
        }}
        
        function updateAllPlotsVisibility() {{
            const allPlotIds = [
                'mean_plot_0', 'mean_plot_1', 'mean_plot_2', 'mean_plot_3',
                'p90_plot_0', 'p90_plot_1', 'p90_plot_2', 'p90_plot_3'
            ];
            
            allPlotIds.forEach(plotId => {{
                const plotDiv = document.getElementById(plotId);
                if (plotDiv && plotDiv.data) {{
                    const updates = {{}};
                    plotDiv.data.forEach((trace, index) => {{
                        const visible = legendVisibility[trace.name] !== false;
                        updates[`visible[${{index}}]`] = visible;
                    }});
                    
                    Plotly.restyle(plotId, updates);
                }}
            }});
        }}
        
        function updatePlots() {{
            const filteredData = filterData();
            const xAxis = document.getElementById('x_axis_select').value;
            const swapAxes = xAxis === 'request_throughput';
            
            if (filteredData.length === 0) {{
                console.log('No data after filtering');
                return;
            }}
            
            // Initialize legend visibility for new traces
            const traces = createTrace(filteredData, 'max_concurrency', 'mean_ttft_ms');
            traces.forEach(trace => {{
                if (legendVisibility[trace.name] === undefined) {{
                    legendVisibility[trace.name] = true;
                }}
            }});
            
            // Create mean metric plots
            metrics.forEach((metric, i) => {{
                let traces, layout;
                
                if (swapAxes) {{
                    // When X-axis is "request_throughput", swap so metric is X and throughput is Y
                    traces = createTrace(filteredData, metric[0], 'request_throughput');
                    layout = {{
                        title: `Request Throughput vs ${{metric[1]}}`,
                        xaxis: {{ title: metric[1] }},
                        yaxis: {{ title: 'Request Throughput (req/s)' }},
                        showlegend: i === 0,
                        legend: {{
                            x: 0.5,
                            y: -0.2,
                            xanchor: 'center',
                            yanchor: 'top',
                            orientation: 'h',
                            bgcolor: 'rgba(255,255,255,0.8)',
                            bordercolor: 'rgba(0,0,0,0.2)',
                            borderwidth: 1
                        }},
                        margin: {{ t: 50, r: 50, b: 100, l: 80 }}
                    }};
                }} else {{
                    // Normal case: concurrency is X, metric is Y
                    traces = createTrace(filteredData, xAxis, metric[0]);
                    layout = {{
                        title: `${{metric[1]}} vs Concurrency`,
                        xaxis: {{ title: 'Concurrency' }},
                        yaxis: {{ title: metric[1] }},
                        showlegend: i === 0,
                        legend: {{
                            x: 0.5,
                            y: -0.2,
                            xanchor: 'center',
                            yanchor: 'top',
                            orientation: 'h',
                            bgcolor: 'rgba(255,255,255,0.8)',
                            bordercolor: 'rgba(0,0,0,0.2)',
                            borderwidth: 1
                        }},
                        margin: {{ t: 50, r: 50, b: 100, l: 80 }}
                    }};
                }}
                
                Plotly.newPlot(`mean_plot_${{i}}`, traces, layout, {{responsive: true}}).then(function() {{
                    // Setup legend click handler only for the plot with legend after plot is created
                    if (i === 0) {{
                        document.getElementById(`mean_plot_${{i}}`).on('plotly_legendclick', function(data) {{
                            const traceName = data.data[data.curveNumber].name;
                            
                            // Toggle visibility state
                            if (legendVisibility[traceName] === undefined) {{
                                legendVisibility[traceName] = true;
                            }}
                            legendVisibility[traceName] = !legendVisibility[traceName];
                            
                            // Update all plots with new visibility
                            updateAllPlotsVisibility();
                            
                            // Prevent default legend behavior
                            return false;
                        }});
                    }}
                }});
            }});
            
            // Create P90 metric plots  
            p90_metrics.forEach((metric, i) => {{
                let traces, layout;
                
                if (swapAxes) {{
                    // When X-axis is "request_throughput", swap so metric is X and throughput is Y
                    traces = createTrace(filteredData, metric[0], 'request_throughput');
                    layout = {{
                        title: `Request Throughput vs ${{metric[1]}}`,
                        xaxis: {{ title: metric[1] }},
                        yaxis: {{ title: 'Request Throughput (req/s)' }},
                        showlegend: false,
                        margin: {{ t: 50, r: 50, b: 50, l: 80 }}
                    }};
                }} else {{
                    // Normal case: concurrency is X, metric is Y
                    traces = createTrace(filteredData, xAxis, metric[0]);
                    layout = {{
                        title: `${{metric[1]}} vs Concurrency`,
                        xaxis: {{ title: 'Concurrency' }},
                        yaxis: {{ title: metric[1] }},
                        showlegend: false,
                        margin: {{ t: 50, r: 50, b: 50, l: 80 }}
                    }};
                }}
                
                Plotly.newPlot(`p90_plot_${{i}}`, traces, layout, {{responsive: true}});
            }});
        }}
        
        // Initialize plots when page loads
        document.addEventListener('DOMContentLoaded', function() {{
            updatePlots();
        }});
    </script>
</body>
</html>
"""
    
    # Write HTML file
    html_file = os.path.join(plot_dir, "interactive_benchmark_dashboard.html")
    with open(html_file, 'w') as f:
        f.write(html_content)
    
    print(f"Interactive dashboard created: {html_file}")
    return html_file

def recalculate_ttfs(df, first_sentence_len):
    if 'mean_ttfs_ms' in df.columns and 'mean_ttft_ms' in df.columns and 'mean_tpot_ms' in df.columns :
        df['mean_ttfs_ms'] = df['mean_ttft_ms'] + df['mean_tpot_ms'] * first_sentence_len
            
    if 'p90_ttfs_ms' in df.columns and 'p90_ttft_ms' in df.columns and 'p90_tpot_ms' in df.columns :
        df['p90_ttfs_ms'] = df['p90_ttft_ms'] + df['p90_tpot_ms'] * first_sentence_len
            
    if 'std_ttfs_ms' in df.columns and 'std_ttft_ms' in df.columns and 'std_tpot_ms' in df.columns :
            # This is an approximation for std of sum of dependent variables
        df['std_ttfs_ms'] = np.sqrt(df['std_ttft_ms']**2 + (df['std_tpot_ms'] * first_sentence_len)**2)


import json
import os
import sys
import time
import traceback
from dataclasses import dataclass, field
from typing import List, Optional, Union
from .util import check_first_sentence

import aiohttp
from tqdm.asyncio import tqdm
AIOHTTP_TIMEOUT = aiohttp.ClientTimeout(total=6 * 60 * 60)


@dataclass
class RequestFuncInput:
    prompt: str
    api_url: str
    prompt_len: int
    output_len: int
    model: str
    best_of: int = 1
    logprobs: Optional[int] = None
    extra_body: Optional[dict] = None
    multi_modal_content: Optional[dict] = None
    ignore_eos: bool = False


@dataclass
class RequestFuncOutput:
    generated_text: str = ""
    first_sentence_text: str = ""
    success: bool = False
    latency: float = 0.0
    ttft: float = 0.0  # Time to first token
    ttfs: float = 0.0 # Time to first sentence
    itl: List[float] = field(
        default_factory=list)  # List of inter-token latencies
    tpot: float = 0.0  # avg next-token latencies
    prompt_len: int = 0
    error: str = ""
    output_tokens: List[int] = field(default_factory=list),
    output_token_len: int = 0
    input_token_len: int = 0


async def async_request_tgi(
    request_func_input: RequestFuncInput,
    pbar: Optional[tqdm] = None,
) -> RequestFuncOutput:
    api_url = request_func_input.api_url
    assert api_url.endswith("generate_stream")

    async with aiohttp.ClientSession(timeout=AIOHTTP_TIMEOUT) as session:
        params = {
            "best_of": request_func_input.best_of,
            "max_new_tokens": request_func_input.output_len,
            "do_sample": True,
            "temperature": 0.01,  # TGI does not accept 0.0 temperature.
            "top_p": 0.99,  # TGI does not accept 1.0 top_p.
            "truncate": request_func_input.prompt_len,
            # TGI does not accept ignore_eos flag.
        }
        payload = {
            "inputs": request_func_input.prompt,
            "parameters": params,
        }
        output = RequestFuncOutput()
        output.prompt_len = request_func_input.prompt_len

        ttft = 0.0
        st = time.perf_counter()
        most_recent_timestamp = st
        try:
            async with session.post(url=api_url, json=payload) as response:
                if response.status == 200:
                    async for chunk_bytes in response.content:
                        chunk_bytes = chunk_bytes.strip()
                        if not chunk_bytes:
                            continue
                        chunk_bytes = chunk_bytes.decode("utf-8")

                        #NOTE: Sometimes TGI returns a ping response without
                        # any data, we should skip it.
                        if chunk_bytes.startswith(":"):
                            continue
                        chunk = chunk_bytes.removeprefix("data:")

                        data = json.loads(chunk)
                        timestamp = time.perf_counter()
                        # First token
                        if ttft == 0.0:
                            ttft = time.perf_counter() - st
                            output.ttft = ttft

                        # Decoding phase
                        else:
                            output.itl.append(timestamp -
                                              most_recent_timestamp)

                        most_recent_timestamp = timestamp

                    output.latency = most_recent_timestamp - st
                    output.success = True
                    output.generated_text = data["generated_text"]
                else:
                    output.error = response.reason or ""
                    output.success = False
        except Exception:
            output.success = False
            exc_info = sys.exc_info()
            output.error = "".join(traceback.format_exception(*exc_info))

        if pbar:
            pbar.update(1)
        return output


async def async_request_trt_llm(
    request_func_input: RequestFuncInput,
    pbar: Optional[tqdm] = None,
) -> RequestFuncOutput:
    api_url = request_func_input.api_url
    assert api_url.endswith("generate_stream")

    async with aiohttp.ClientSession(timeout=AIOHTTP_TIMEOUT) as session:
        assert request_func_input.best_of == 1
        payload = {
            "accumulate_tokens": True,
            "text_input": request_func_input.prompt,
            "temperature": 0.0,
            "top_p": 1.0,
            "max_tokens": request_func_input.output_len,
            "stream": True,
        }
        if request_func_input.ignore_eos:
            payload["min_length"] = request_func_input.output_len
        output = RequestFuncOutput()
        output.prompt_len = request_func_input.prompt_len

        ttft = 0.0
        st = time.perf_counter()
        most_recent_timestamp = st
        try:
            async with session.post(url=api_url, json=payload) as response:
                if response.status == 200:
                    async for chunk_bytes in response.content:
                        chunk_bytes = chunk_bytes.strip()
                        if not chunk_bytes:
                            continue

                        chunk = chunk_bytes.decode("utf-8").removeprefix(
                            "data:")

                        data = json.loads(chunk)
                        output.generated_text += data["text_output"]
                        timestamp = time.perf_counter()
                        # First token
                        if ttft == 0.0:
                            ttft = time.perf_counter() - st
                            output.ttft = ttft

                        # Decoding phase
                        else:
                            output.itl.append(timestamp -
                                              most_recent_timestamp)

                        most_recent_timestamp = timestamp

                    output.latency = most_recent_timestamp - st
                    output.success = True

                else:
                    output.error = response.reason or ""
                    output.success = False
        except Exception:
            output.success = False
            exc_info = sys.exc_info()
            output.error = "".join(traceback.format_exception(*exc_info))

        if pbar:
            pbar.update(1)
        return output


async def async_request_deepspeed_mii(
    request_func_input: RequestFuncInput,
    pbar: Optional[tqdm] = None,
) -> RequestFuncOutput:
    async with aiohttp.ClientSession(timeout=AIOHTTP_TIMEOUT) as session:
        assert request_func_input.best_of == 1

        payload = {
            "prompt": request_func_input.prompt,
            "max_tokens": request_func_input.output_len,
            "temperature": 0.01,  # deepspeed-mii does not accept 0.0 temp.
            "top_p": 1.0,
        }
        output = RequestFuncOutput()
        output.prompt_len = request_func_input.prompt_len

        # NOTE: DeepSpeed-MII doesn't support streaming as of Jan 28 2024,
        # will use 0 as placeholder.
        # See https://github.com/microsoft/DeepSpeed-MII/pull/311
        output.ttft = 0

        st = time.perf_counter()
        try:
            async with session.post(url=request_func_input.api_url,
                                    json=payload) as response:
                if response.status == 200:
                    parsed_resp = await response.json()
                    output.latency = time.perf_counter() - st
                    output.generated_text = parsed_resp["text"][0]
                    output.success = True
                else:
                    output.error = response.reason or ""
                    output.success = False
        except Exception:
            output.success = False
            exc_info = sys.exc_info()
            output.error = "".join(traceback.format_exception(*exc_info))

        if pbar:
            pbar.update(1)
        return output


async def async_request_openai_completions(
    request_func_input: RequestFuncInput,
    pbar: Optional[tqdm] = None,
) -> RequestFuncOutput:
    api_url = request_func_input.api_url
    assert api_url.endswith(
        ("completions", "profile")
    ), "OpenAI Completions API URL must end with 'completions' or 'profile'."

    async with aiohttp.ClientSession(timeout=AIOHTTP_TIMEOUT) as session:
        if isinstance(request_func_input.prompt, list):
            request_func_input.prompt = [int(token) for token in request_func_input.prompt]

        payload = {
            "model": request_func_input.model,
            "prompt": request_func_input.prompt,
            "temperature": 0.0,
            "best_of": request_func_input.best_of,
            "max_tokens": request_func_input.output_len,
            "logprobs": request_func_input.logprobs,
            "stream": True,
            "stream_options": {"include_usage": True, "continuous_usage_stats": True},
            "ignore_eos": request_func_input.ignore_eos,
        }
        if request_func_input.extra_body:
            payload.update(request_func_input.extra_body)
        headers = {
            "Authorization": f"Bearer {os.environ.get('OPENAI_API_KEY')}"
        }

        output = RequestFuncOutput()
        output.prompt_len = request_func_input.prompt_len

        generated_text = ""
        ttft = 0.0
        st = time.perf_counter()
        most_recent_timestamp = st
        
        # for ttfs
        first_sentence_time = 0.0
        start_fs_flag = 0
        first_sentence_text = ""
        try:
            async with session.post(url=api_url, json=payload,
                                    headers=headers) as response:
                if response.status == 200:
                    first_chunk_received = False
                    async for chunk_bytes in response.content:
                        chunk_bytes = chunk_bytes.strip()
                        if not chunk_bytes:
                            continue

                        chunk = chunk_bytes.decode("utf-8").removeprefix(
                            "data: ")
                        # import pdb; pdb.set_trace()
                        if chunk == "[DONE]":
                            latency = time.perf_counter() - st
                        else:
                            data = json.loads(chunk)

                            # NOTE: Some completion API might have a last
                            # usage summary response without a token so we
                            # want to check a token was generated

                            if data["choices"] and data["choices"][0] and data["choices"][0]["text"]:

                                timestamp = time.perf_counter()
                                # First token
                                if not first_chunk_received:
                                    first_chunk_received = True
                                    ttft = time.perf_counter() - st
                                    output.ttft = ttft

                                # Decoding phase
                                else:
                                    output.itl.append(timestamp -
                                                      most_recent_timestamp)
                                    # collect the token information for speculative decoding
                                    if data["choices"][0]["logprobs"]:
                                        output.output_tokens += data["choices"][0]["logprobs"]['tokens']
                                most_recent_timestamp = timestamp
                                cur_text = data["choices"][0]["text"]
                                
                                # handle ttfs
                                if start_fs_flag and cur_text is not None and (
                                        '[' in cur_text or '"' in cur_text) and first_sentence_time == 0.:
                                    first_sentence_time = time.perf_counter() - st
                                if not start_fs_flag and cur_text is not None and '[' in cur_text:
                                    start_fs_flag = True
                                if first_sentence_time == 0.:
                                    first_sentence_text += cur_text
                                
                                generated_text += cur_text

                    if first_chunk_received:
                        output.success = True
                    else:
                        output.success = False
                        output.error = (
                            "Never received a valid chunk to calculate TTFT."
                            "This response will be marked as failed!")
                    # we find sometimes server side will output empty string, which will not be stream back to client
                    # thus relying on usage information from server is more reliable
                    if data['usage']:
                        output.output_token_len = data['usage']['completion_tokens']
                    output.generated_text = generated_text
                    output.first_sentence_text = first_sentence_text
                    output.ttfs = first_sentence_time
                    output.latency = latency
                    # import pdb; pdb.set_trace()
                else:
                    output.error = response.reason or ""
                    output.error += f": request info: api_url: {api_url}, payload: {payload}, headers: {headers}"
                    output.success = False
        except Exception:
            output.success = False
            exc_info = sys.exc_info()
            output.error = "".join(traceback.format_exception(*exc_info))

    if pbar:
        pbar.update(1)
    return output


async def async_request_openai_chat_completions(
    request_func_input: RequestFuncInput,
    pbar: Optional[tqdm] = None,
) -> RequestFuncOutput:
    api_url = request_func_input.api_url
    assert api_url.endswith(
        "chat/completions"
    ), "OpenAI Chat Completions API URL must end with 'chat/completions'."

    async with aiohttp.ClientSession(timeout=AIOHTTP_TIMEOUT) as session:
        content = [{"type": "text", "text": request_func_input.prompt}]
        if request_func_input.multi_modal_content:
            content.append(request_func_input.multi_modal_content)
        payload = {
            "model": request_func_input.model,
            "messages": request_func_input.prompt,
            "temperature": 0.0,
            "max_tokens": request_func_input.output_len,
            "stream": True,
            "ignore_eos": request_func_input.ignore_eos,
            "logprobs": request_func_input.logprobs,
            "stream_options": {"include_usage": True, "continuous_usage_stats": True},
        }
        if request_func_input.extra_body:
            payload.update(request_func_input.extra_body)
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {os.environ.get('OPENAI_API_KEY')}",
        }

        output = RequestFuncOutput()
        output.prompt_len = request_func_input.prompt_len
        text_list=[]

        generated_text = ""
        ttft = 0.0
        st = time.perf_counter()
        most_recent_timestamp = st


        # for ttfs
        first_sentence_time = 0.0
        start_fs_flag = 0
        first_sentence_text = ""
        previous_len = 0
        try:
            async with session.post(url=api_url, json=payload,
                                    headers=headers) as response:
                if response.status == 200:
                    async for chunk_bytes in response.content:
                        chunk_bytes = chunk_bytes.strip()
                        if not chunk_bytes:
                            continue
                        chunk = chunk_bytes.decode("utf-8").removeprefix(
                            "data: ")
                        if chunk == "[DONE]":
                            latency = time.perf_counter() - st
                        else:
                            timestamp = time.perf_counter()
                            data = json.loads(chunk)
                            # print("data", data)
                            # usge information is usually included in the last chunk
                            if "usage" in data:
                                output.output_token_len = data["usage"]["completion_tokens"]
                                output.input_token_len = data["usage"]["prompt_tokens"]

                            if not data["choices"]:
                                continue
                            delta = data["choices"][0]["delta"]
                            # print(f'data["choices"][0]["delta"]["content"]')
                            if delta.get("content", None):
                                # First token
                                if ttft == 0.0:
                                    ttft = time.perf_counter() - st
                                    output.ttft = ttft

                                # Decoding phase
                                else:
                                    output.itl.append(timestamp -
                                                      most_recent_timestamp)

                                cur_text = delta["content"]
                                # text_list.append([cur_text, data["usage"]["completion_tokens"]-previous_len])
                                # previous_len = data["usage"]["completion_tokens"]
                                # print(text_list)
                                generated_text += cur_text
                                # handle ttfs
                                # old first sentence logic:
                                # if start_fs_flag and cur_text is not None and (
                                #         '[' in cur_text or '"' in cur_text) and first_sentence_time == 0.:
                                #     first_sentence_time = time.perf_counter() - 
                                # if not start_fs_flag and cur_text is not None and '[' in cur_text:
                                #     start_fs_flag = True
                                # if first_sentence_time == 0.:
                                #     first_sentence_text += cur_text


                                # new first sentence logic:
                                if start_fs_flag and first_sentence_time == 0:

                                    if check_first_sentence(generated_text):
                                        first_sentence_time = time.perf_counter() - st
                                        first_sentence_text = generated_text
                                if not start_fs_flag and '[' in cur_text:
                                    start_fs_flag = True

                            most_recent_timestamp = timestamp


                    output.generated_text = generated_text
                    try:
                        json.loads(generated_text)
                        output.success = True
                    except json.JSONDecodeError:
                        output.success = False
                        output.error = "Invalid JSON response"
                        print(generated_text)
                    output.latency = latency
                    output.ttfs = first_sentence_time
                    output.first_sentence_text = first_sentence_text
                    # import pdb; pdb.set_trace()
                else:
                    output.error = response.reason or ""
                    output.success = False
        except Exception:
            output.success = False
            exc_info = sys.exc_info()
            output.error = "".join(traceback.format_exception(*exc_info))

    if pbar:
        pbar.update(1)
    return output


def get_model(pretrained_model_name_or_path: str) -> str:
    if os.getenv('VLLM_USE_MODELSCOPE', 'False').lower() == 'true':
        from modelscope import snapshot_download
        import huggingface_hub.constants
        model_path = snapshot_download(
            model_id=pretrained_model_name_or_path,
            local_files_only=huggingface_hub.constants.HF_HUB_OFFLINE,
            ignore_file_pattern=[".*.pt", ".*.safetensors", ".*.bin"])

        return model_path
    return pretrained_model_name_or_path


def get_tokenizer(
    pretrained_model_name_or_path: str, trust_remote_code: bool
) :
    from transformers import AutoTokenizer

    if pretrained_model_name_or_path is not None and not os.path.exists(
            pretrained_model_name_or_path):
        pretrained_model_name_or_path = get_model(
            pretrained_model_name_or_path)
    return AutoTokenizer.from_pretrained(pretrained_model_name_or_path,
                                         trust_remote_code=trust_remote_code)


ASYNC_REQUEST_FUNCS = {
    "tgi": async_request_tgi,
    "vllm": async_request_openai_completions,
    "lmdeploy": async_request_openai_completions,
    "deepspeed-mii": async_request_deepspeed_mii,
    "openai": async_request_openai_completions,
    "openai-chat": async_request_openai_chat_completions,
    "tensorrt-llm": async_request_trt_llm,
    "scalellm": async_request_openai_completions,
    "sglang": async_request_openai_completions,
}





#!/usr/bin/env python3
import subprocess
import requests
import json
import numpy as np
import time
import os
import tempfile
from pathlib import Path
import signal
import sys
import threading
from opentelemetry import trace
from opentelemetry.trace import TracerProvider

# Configuration
VLLM_SERVER = "http://localhost:8004"  # Your vLLM server endpoint
NUM_REQUESTS = 10                     # Number of requests to send
PROMPT = "Write a short poem about AI"  # Prompt to send to vLLM
MAX_TOKENS = 20                       # Max tokens to generate

def setup_otel_collector(server_metric_path):
    """Setup and start the OpenTelemetry Collector"""
    # Create config file
    config = """
receivers:
      otlp:
        protocols:
          grpc: # Use gRPC
            endpoint: 0.0.0.0:4317
        # or use a completely different port if needed:
        # endpoint: localhost:5318

exporters:
  file:
    path: /mnt/personal/honbgo_vllm/graphana/jaeger/jaeger-1.57.0-linux-amd64/tmp/latency_data.json
    format: json
  logging:
    verbosity: detailed # Use detailed for max info

processors:
  batch:
    timeout: 1s

service:
  telemetry: # Keep your telemetry config if you added it
        metrics:
          address: :8889

  pipelines:
    traces:
      receivers: [otlp]
      processors: [batch]
      exporters: [file, logging]

"""
    
    config_path = Path("/tmp/otel-config.yaml")
    with open(config_path, "w") as f:
        f.write(config)
    
    # Download collector if not exists
    collector_path = Path("/tmp/otelcol-contrib")
    if not collector_path.exists():
        print("Downloading OpenTelemetry Collector...")
        try:
            subprocess.run(
                "curl -L https://github.com/open-telemetry/opentelemetry-collector-releases/releases/download/v0.103.1/otelcol-contrib_0.103.1_linux_amd64.tar.gz -o /tmp/otelcol.tar.gz",
                shell=True, check=True
            )
            subprocess.run("tar -xzf /tmp/otelcol.tar.gz -C /tmp", shell=True, check=True)
            
            # Create directory for the output file if it doesn't exist
            subprocess.run("chmod +x /tmp/otelcol-contrib", shell=True, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error downloading collector: {e}")
            sys.exit(1)
    
    # Start collector
    print("Starting OpenTelemetry Collector...")
    try:
        process = subprocess.Popen(
            ["/tmp/otelcol-contrib", "--config", "/tmp/otel-config.yaml", "--set", f"exporters.file.path={server_metric_path}"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        
        # Give it a moment to start
        time.sleep(2)
        
        # Check if it's running
        if process.poll() is not None:
            out, err = process.communicate()
            print(f"Collector failed to start: {err.decode('utf-8')}")
            sys.exit(1)
            
        # Add these lines to capture output in real-time
        def log_output(process):
            while True:
                line = process.stderr.readline()
                if not line:
                    break
                print(f"COLLECTOR: {line.decode().strip()}")
        
        threading.Thread(target=log_output, args=(process,), daemon=True).start()
        
        return process
    except Exception as e:
        print(f"Error starting collector: {e}")
        sys.exit(1)

def send_requests_to_vllm():
    """Send requests to the vLLM server and return completion times"""
    print(f"Sending {NUM_REQUESTS} requests to vLLM server...")
    
    for i in range(NUM_REQUESTS):
        try:
            response = requests.post(
                f"{VLLM_SERVER}/v1/completions",
                json={
                    "model": "/mnt/project/llm/ckpt/stable_ckpts/Llama-3.2-1B/",  # Model name - adjust if needed
                    "prompt": PROMPT,
                    "max_tokens": MAX_TOKENS,
                    "temperature": 0.7
                },
                headers={"Content-Type": "application/json"}
            )
            
            if response.status_code == 200:
                print(f"Request {i+1}/{NUM_REQUESTS} completed successfully")
            else:
                print(f"Request {i+1}/{NUM_REQUESTS} failed: {response.status_code} - {response.text}")
                
            # After your request code
            print(response.json())
            
        except Exception as e:
            print(f"Error sending request: {e}")
        
        # Small delay between requests
        time.sleep(0.5)

    # Add this cleanup code

def calculate_percentiles(service_metric_path):
    """Process the trace data from the specified file and calculate percentiles"""
    print(f"Attempting to read trace data from: {service_metric_path}")
    trace_file = Path(service_metric_path)

    # Give the file system and collector flush interval a bit more time
    time.sleep(7) # Increased wait time slightly after finishing requests

    if not trace_file.exists():
        print(f"Error: Trace data file not found: {service_metric_path}", file=sys.stderr)
        return

    if trace_file.stat().st_size == 0:
        print(f"Error: Trace data file is empty: {service_metric_path}", file=sys.stderr)
        return

    latencies = []
    total_spans_processed = 0
    lines_processed = 0
    trace_data_found = False

    try:
        with open(trace_file, 'r') as f:
            # Try reading as JSON Lines first (more likely with streaming writes)
            print("Attempting to read as JSON Lines...")
            f.seek(0) # Ensure we start at the beginning
            
            # Track multiple latency metrics
            latency_metrics = {
                'gen_ai.latency.time_to_first_token': [],
                'gen_ai.latency.e2e': [],
                'gen_ai.latency.time_in_queue': [],
                'gen_ai.latency.time_in_scheduler': [],
                'gen_ai.latency.time_in_model_forward': [],
                'gen_ai.latency.time_in_model_execute': [],
                'gen_ai.running_batch_size': [],
                'gen_ai.running_batch_size_list': [],
                'gen_ai.latency.time_in_model_forward_list': [],
                'gen_ai.latency.time_in_model_execute_list': []
            }
            
            for line_num, line in enumerate(f):
                lines_processed = line_num + 1
                line = line.strip()
                if not line:
                    continue
                try:
                    # Each line should contain a ResourceSpans object or similar structure
                    data_line = json.loads(line)
                    trace_data_found = True # Mark that we found at least one valid JSON line

                    # Check if the line contains the expected structure
                    if isinstance(data_line, dict):
                        resource_spans_list = data_line.get('resourceSpans', [])
                        # Sometimes the root object itself is the ResourceSpans object
                        if not resource_spans_list and "resource" in data_line and "scopeSpans" in data_line:
                             resource_spans_list = [data_line] # Treat the line object as a single ResourceSpans

                        for resource_span in resource_spans_list:
                            for scope_spans in resource_span.get('scopeSpans', []):
                                for span in scope_spans.get('spans', []):
                                    total_spans_processed += 1
                                    # Check for all latency attributes
                                    for attribute in span.get('attributes', []):
                                        attr_key = attribute.get('key')
                                        if attr_key in latency_metrics:
                                            value_info = attribute.get('value', {})
                                            if 'doubleValue' in value_info:
                                                latency_metrics[attr_key].append(float(value_info['doubleValue']))
                                            elif 'intValue' in value_info: # Handle potential int values too
                                                latency_metrics[attr_key].append(float(value_info['intValue']))
                                            # Handle list values (arrays)
                                            elif 'arrayValue' in value_info:
                                                array_values = value_info.get('arrayValue', {}).get('values', [])
                                                current_array = []
                                                for val in array_values:
                                                    if 'doubleValue' in val:
                                                        current_array.append(float(val['doubleValue']))
                                                    elif 'intValue' in val:
                                                        current_array.append(float(val['intValue']))
                                                if current_array:  # Only append if we have values
                                                    latency_metrics[attr_key].append(current_array)

                except json.JSONDecodeError as line_e:
                    print(f"Warning: Could not decode line {line_num + 1} as JSON: {line_e}. Content: '{line[:100]}...'", file=sys.stderr)
                except Exception as line_e:
                     print(f"Warning: Error processing line {line_num + 1}: {line_e}", file=sys.stderr)

            # If JSON Lines didn't yield data or format seemed wrong, maybe it *is* a single object
            if not trace_data_found and lines_processed <= 1:
                 print("Could not parse as JSON Lines or file was very short, trying as single JSON object...")
                 f.seek(0)
                 try:
                     data = json.load(f)
                     lines_processed = 1
                     if isinstance(data, dict) and 'resourceSpans' in data:
                         resource_spans_list = data.get('resourceSpans', [])
                         for resource_span in resource_spans_list:
                            for scope_spans in resource_span.get('scopeSpans', []):
                                for span in scope_spans.get('spans', []):
                                    total_spans_processed += 1
                                    for attribute in span.get('attributes', []):
                                        attr_key = attribute.get('key')
                                        if attr_key in latency_metrics:
                                            value_info = attribute.get('value', {})
                                            if 'doubleValue' in value_info:
                                                latency_metrics[attr_key].append(float(value_info['doubleValue']))
                                            elif 'intValue' in value_info:
                                                latency_metrics[attr_key].append(float(value_info['intValue']))
                                            # Handle list values (arrays) as 2D array
                                            elif 'arrayValue' in value_info:
                                                array_values = value_info.get('arrayValue', {}).get('values', [])
                                                current_array = []
                                                for val in array_values:
                                                    if 'doubleValue' in val:
                                                        current_array.append(float(val['doubleValue']))
                                                    elif 'intValue' in val:
                                                        current_array.append(float(val['intValue']))
                                                import pdb; pdb.set_trace()
                                                if current_array:  # Only append if we have values
                                                    latency_metrics[attr_key].append(current_array)
                         trace_data_found = True # Mark that we read the single object
                 except json.JSONDecodeError as single_e:
                     print(f"Error: Could not decode file as single JSON object either: {single_e}", file=sys.stderr)
                 except Exception as single_e:
                    print(f"Error processing file as single JSON object: {single_e}", file=sys.stderr)


    except FileNotFoundError:
        print(f"Error: Trace data file disappeared during processing: {service_metric_path}", file=sys.stderr)
        return
    except Exception as e:
        print(f"Error reading or processing trace data file: {e}", file=sys.stderr)
        return

    print(f"Processed {lines_processed} JSON line(s)/object(s). Found {total_spans_processed} total spans.")

    # For backward compatibility
    latencies = latency_metrics['gen_ai.latency.time_to_first_token']

    # Calculate percentiles for all metrics
    try:
        import numpy as np
    except ImportError:
        print("Error: numpy is required for percentile calculations. Please install it (`pip install numpy`).", file=sys.stderr)
        return

    # Print statistics for each latency metric
    for metric_name, values in latency_metrics.items():
        if values:
            print(metric_name, values)
            p50 = np.percentile(values, 50)
            p90 = np.percentile(values, 90)
            p95 = np.percentile(values, 95)
            p99 = np.percentile(values, 99)

            print(f"\n=== {metric_name} Statistics ===")
            print(f"P50: {p50:.6f}s")
            print(f"P90: {p90:.6f}s")
            print(f"P95: {p95:.6f}s")
            print(f"P99: {p99:.6f}s")
            print(f"Min: {min(values):.6f}s")
            print(f"Max: {max(values):.6f}s")
            print(f"Mean: {np.mean(values):.6f}s")
            print(f"Std Dev: {np.std(values):.6f}s")
            print(f"Total samples: {len(values)}")
        elif trace_data_found:
            print(f"Found trace data but no spans contained the '{metric_name}' attribute.")
    
    if not trace_data_found:
        print(f"No valid trace data found in the file to extract metrics from.")

def main():
    # Start collector
    server_metric_path = "/mnt/personal/honbgo_vllm/graphana/latency_data.json"
    collector_process = setup_otel_collector(server_metric_path)
    
    try:
        # Send requests
        send_requests_to_vllm()
        
        # Process results
        calculate_percentiles(server_metric_path)
    finally:
        # Cleanup
        print("Shutting down collector...")
        if collector_process:
            collector_process.terminate()
            collector_process.wait()

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
vLLM Metrics Analyzer

This script collects, analyzes, and calculates percentiles from vLLM's Prometheus metrics.
It can be used either as a standalone script or imported by other modules.
"""

import argparse
import json
import sys
import time
from collections import defaultdict
from typing import Dict, List, Optional, Any, Union, Tuple

import requests
from prometheus_client.parser import text_string_to_metric_families

# Default configuration
DEFAULT_VLLM_SERVER = "http://localhost:8004"
DEFAULT_METRICS_ENDPOINT = "/metrics"
DEFAULT_MODEL_NAME = None  # Will use server's default
DEFAULT_PROMPT = "Write a short poem about artificial intelligence."
DEFAULT_MAX_TOKENS = 20
DEFAULT_NUM_REQUESTS = 10
DEFAULT_PERCENTILES = [1, 25, 50, 75, 99]


def collect_prometheus_metrics(server_url: str = DEFAULT_VLLM_SERVER,
                     metrics_path: str = DEFAULT_METRICS_ENDPOINT,
                     save_raw_path: str = "./raw_prometheus_metrics.txt") -> Optional[str]:
    """
    Collect raw metrics from vLLM's Prometheus metrics endpoint.

    Args:
        server_url: Base URL of the vLLM server
        metrics_path: Path to the metrics endpoint
        save_raw_path: Optional path to save raw metrics text

    Returns:
        Raw metrics text or None if request fails
    """
    metrics_url = f"{server_url.rstrip('/')}{metrics_path}"
    
    try:
        print(f"Fetching metrics from {metrics_url}...")
        response = requests.get(metrics_url, timeout=10)
        response.raise_for_status()
        
        metrics_text = response.text
        
        # Save raw metrics to file if path is specified
        if save_raw_path:
            try:
                with open(save_raw_path, 'w') as f:
                    f.write(metrics_text)
                print(f"Raw metrics saved to {save_raw_path}")
            except Exception as e:
                print(f"Error saving raw metrics: {e}", file=sys.stderr)
        
        return metrics_text
    except requests.RequestException as e:
        print(f"Error fetching metrics: {e}", file=sys.stderr)
        return None


def send_requests_to_vllm(server_url: str = DEFAULT_VLLM_SERVER,
                         num_requests: int = DEFAULT_NUM_REQUESTS,
                         model_name: Optional[str] = DEFAULT_MODEL_NAME,
                         prompt: str = DEFAULT_PROMPT,
                         max_tokens: int = DEFAULT_MAX_TOKENS) -> bool:
    """
    Send test requests to the vLLM server to generate metrics.

    Args:
        server_url: Base URL of the vLLM server
        num_requests: Number of requests to send
        model_name: Model name to use (None for server default)
        prompt: Prompt text to send
        max_tokens: Maximum tokens to generate

    Returns:
        True if all requests were successfully sent, False otherwise
    """
    success_count = 0
    completions_url = f"{server_url.rstrip('/')}/v1/completions"
    
    print(f"Sending {num_requests} test requests to {completions_url}...")
    
    for i in range(num_requests):
        try:
            payload = {
                "prompt": prompt,
                "max_tokens": max_tokens,
                "temperature": 0,
                "top_k": 1,
                "ignore_eos": True
            }
            
            # Add model name if specified
            if model_name:
                payload["model"] = model_name
                
            response = requests.post(
                completions_url,
                json=payload,
                headers={"Content-Type": "application/json"},
                timeout=30
            )
            
            if response.status_code == 200:
                print(f"Request {i+1}/{num_requests} completed successfully")
                success_count += 1
            else:
                print(f"Request {i+1}/{num_requests} failed: {response.status_code} - {response.text}",
                     file=sys.stderr)
                
        except Exception as e:
            print(f"Error sending request {i+1}/{num_requests}: {e}", file=sys.stderr)
        
        # Small delay between requests
        if i < num_requests - 1:
            time.sleep(0.5)
    
    success_rate = success_count / num_requests if num_requests > 0 else 0
    print(f"Completed {success_count}/{num_requests} requests ({success_rate:.1%} success rate)")
    
    return success_count == num_requests


def analyze_metrics(metrics_text: str, 
                   percentiles: List[int] = DEFAULT_PERCENTILES,
                   save_results_path: str = "./processed_prometheus_metrics.json") -> Dict[str, Dict[str, Any]]:
    """
    Analyze vLLM metrics and calculate specified percentiles.

    Args:
        metrics_text: Raw Prometheus metrics text
        percentiles: List of percentiles to calculate (1-99)
        save_results_path: Optional path to save analysis results as JSON

    Returns:
        Dictionary of analyzed metrics with percentiles
    """
    if not metrics_text:
        return {}
    
    # Organize metrics by type
    metrics_by_type = {
        'histograms': defaultdict(lambda: {'buckets': {}, 'sum': 0, 'count': 0}),
        'counters': {},
        'gauges': {}
    }
    
    # Parse metrics
    for family in text_string_to_metric_families(metrics_text):
        family_name = family.name
        family_type = getattr(family, 'type', 'unknown')
        
        if family_type == 'histogram':
            # Process histogram metrics
            for sample in family.samples:
                sample_name = sample.name
                
                if sample_name.endswith('_sum'):
                    metrics_by_type['histograms'][family_name]['sum'] = sample.value
                elif sample_name.endswith('_count'):
                    metrics_by_type['histograms'][family_name]['count'] = sample.value
                elif sample_name.endswith('_bucket'):
                    if 'le' in sample.labels:
                        try:
                            # Skip +Inf bucket
                            if sample.labels['le'] == '+Inf':
                                continue
                            bucket_value = float(sample.labels['le'])
                            metrics_by_type['histograms'][family_name]['buckets'][bucket_value] = sample.value
                        except (ValueError, TypeError):
                            pass
        
        elif family_type == 'counter':
            # Process counter metrics
            for sample in family.samples:
                metrics_by_type['counters'][sample.name] = sample.value
        
        elif family_type == 'gauge':
            # Process gauge metrics
            for sample in family.samples:
                metrics_by_type['gauges'][sample.name] = sample.value
    
    # Calculate histogram statistics and percentiles
    results = {}
    
    for hist_name, hist_data in metrics_by_type['histograms'].items():
        if hist_data['count'] > 0 and hist_data['buckets']:
            buckets = sorted(hist_data['buckets'].items())
            count = hist_data['count']
            sum_value = hist_data['sum']
            
            stats = {
                'count': count,
                'sum': sum_value,
                'avg': sum_value / count,
                'min': buckets[0][0] if any(v > 0 for _, v in buckets) else None,
                'max': buckets[-1][0]
            }
            # Calculate requested percentiles
            for p in percentiles:
                if p < 1 or p > 99:
                    continue  # Skip invalid percentiles
                
                target_count = count * (p / 100.0)
                percentile_found = False
                
                for i, (upper_bound, cum_count) in enumerate(buckets):
                    if cum_count >= target_count:
                        if i > 0:
                            # Linear interpolation if possible
                            lower_bound, lower_count = buckets[i-1]
                            if cum_count > lower_count:  # Avoid division by zero
                                fraction = (target_count - lower_count) / (cum_count - lower_count)
                                interpolated = lower_bound + fraction * (upper_bound - lower_bound)
                                stats[f'p{p}'] = interpolated
                            else:
                                stats[f'p{p}'] = upper_bound
                        else:
                            stats[f'p{p}'] = upper_bound
                        percentile_found = True
                        break
                
                if not percentile_found:
                    stats[f'p{p}'] = None
            
            results[hist_name] = stats
    
    # Add counter metrics
    for counter_name, value in metrics_by_type['counters'].items():
        results[counter_name] = {'value': value, 'type': 'counter'}
    
    # Add gauge metrics
    for gauge_name, value in metrics_by_type['gauges'].items():
        results[gauge_name] = {'value': value, 'type': 'gauge'}
    
    # Save results to file if path specified
    if save_results_path:
        try:
            with open(save_results_path, 'w') as f:
                json.dump(results, f, indent=2)
            print(f"Analysis results saved to {save_results_path}")
        except Exception as e:
            print(f"Error saving analysis results: {e}", file=sys.stderr)
    print_analysis_results(results)
    return results


def print_analysis_results(results: Dict[str, Dict[str, Any]], 
                          show_histograms: bool = True,
                          show_counters: bool = True,
                          show_gauges: bool = False) -> None:
    """
    Print analysis results in a human-readable format.

    Args:
        results: Analysis results from analyze_metrics()
        show_histograms: Whether to show histogram metrics
        show_counters: Whether to show counter metrics
        show_gauges: Whether to show gauge metrics
    """
    # Organize by metric type
    histograms = {}
    counters = {}
    gauges = {}
    
    for name, data in results.items():
        if 'type' in data:
            if data['type'] == 'counter':
                counters[name] = data
            elif data['type'] == 'gauge':
                gauges[name] = data
        elif 'count' in data and 'avg' in data:
            histograms[name] = data
    
    # Print histograms
    if show_histograms and histograms:
        print("\n=== Prometheus Metrics from vllm ===")
        for name, data in sorted(histograms.items()):
            print(f"\n{name}:")
            print(f"  Count: {data['count']}")
            print(f"  Avg: {data['avg']:.6f}")
            
            # Print all percentiles
            for key in sorted(data.keys()):
                if key.startswith('p') and key[1:].isdigit():
                    value = data[key]
                    if value is not None:
                        print(f"  {key.upper()}: {value:.6f}")
    
    # Print counters
    if show_counters and counters:
        print("\n=== Counter Metrics ===")
        for name, data in sorted(counters.items()):
            print(f"  {name}: {data['value']}")
    
    # Print gauges
    if show_gauges and gauges:
        print("\n=== Gauge Metrics ===")
        for name, data in sorted(gauges.items()):
            print(f"  {name}: {data['value']}")


def main():
    """
    Main function when script is run directly.
    """
    parser = argparse.ArgumentParser(description='Collect and analyze vLLM Prometheus metrics')
    parser.add_argument('--server', default=DEFAULT_VLLM_SERVER, help='vLLM server URL')
    parser.add_argument('--send-requests', action='store_true', help='Send test requests before collecting metrics')
    parser.add_argument('--num-requests', type=int, default=DEFAULT_NUM_REQUESTS, help='Number of test requests to send')
    parser.add_argument('--model',help='Model name to use for requests (server default if not specified)')
    parser.add_argument('--prompt', default=DEFAULT_PROMPT, help='Prompt text to use for requests')
    parser.add_argument('--max-tokens', type=int, default=DEFAULT_MAX_TOKENS, help='Max tokens for generation')
    parser.add_argument('--percentiles', type=str, default='1,25,50,75,99', help='Comma-separated percentiles to calculate')
    parser.add_argument('--output', default = "./processed_prometheus_metrics.json", help='Save analyzed results to JSON file')
    parser.add_argument('--raw-output', default = "./raw_prometheus_metrics.txt", help='Save raw metrics text to file')
    parser.add_argument('--show-gauges', action='store_true', help='Show gauge metrics in output')
    
    args = parser.parse_args()
    if args.send_requests:
        if args.model is None:
            print("Error: --model is required when --send-requests is true", file=sys.stderr)
            return 1
    
    # Parse percentiles
    try:
        percentiles = [int(p.strip()) for p in args.percentiles.split(',')]
    except ValueError:
        print("Error: Percentiles must be comma-separated integers", file=sys.stderr)
        return 1
    
    # Send test requests if requested
    if args.send_requests:
        success = send_requests_to_vllm(
            server_url=args.server,
            num_requests=args.num_requests,
            model_name=args.model,
            prompt=args.prompt,
            max_tokens=args.max_tokens
        )
        if not success:
            print("Warning: Not all requests were successful", file=sys.stderr)
    
    # Collect and analyze metrics
    metrics_text = collect_prometheus_metrics(
        server_url=args.server,
        save_raw_path=args.raw_output
    )
    if not metrics_text:
        print("Error: Failed to collect metrics from vLLM server", file=sys.stderr)
        return 1
    
    analyze_metrics(
        metrics_text, 
        percentiles=percentiles,
        save_results_path=args.output
    )
    
    return 0


if __name__ == "__main__":
    sys.exit(main())



import os
import time

import torch.distributed as dist

from ...utils import get_eval_ckpt_list, invoke_evaluation_service


class AutoEvaluationWorkflowTrigger():
    def __init__(self, 
                 args, 
                 code_info=None):
        self.project = args.eval_project
        self.project_name = args.project_name
        self.modality = args.eval_modality
        self.model_name = args.model_name
        self.cluster = os.getenv("MLP_CLUSTER", "va")
        self.project = os.getenv("MLP_PROJECT", "llm")
        self.train_data_type = args.data_type
        self.train_dataset_name = args.data_type
        self.train_tp = args.tp
        self.train_pp = args.pp
        self.train_cp = args.context_p
        self.train_ep = args.ep
        self.train_sp = args.enable_sp
        self.train_seqlen = args.seq_length
        self.train_batch_size = args.global_batch_size
        self.train_learning_rate = args.lr
        self.train_batch_size = args.global_batch_size
        self.train_tokenizer_path = args.tokenizer_path
        self.eval_tp = args.eval_tp
        self.eval_pp = args.eval_pp
        self.eval_cp = args.eval_cp
        self.eval_ep = args.eval_ep
        self.eval_sp = args.eval_sp
        self.eval_seqlen = args.seq_length  # keep the same as train_seqlen
        self.eval_batch_size = args.eval_batch_size
        self.eval_tokenizer_path = args.tokenizer_path
        # change to eval_dataset_list
        self.eval_dataset_list = get_eval_ckpt_list(
            args.eval_dataset_path, args.eval_dataset_list)
        self.eval_validation_batch_size = args.eval_validation_batch_size
        self.eval_tasks = args.eval_tasks
        self.start_time = time.time()
        self.wandb_project = args.wandb_project
        self.wandb_api_key = os.getenv("WANDB_API_KEY")
        self.inference_max_seq_length = args.inference_max_seq_length
        self.code_info = code_info
        print(f"code_info: {self.code_info}")
        self.train_dataset_name_str = args.dataset_name
        self.train_dataset_ratios_str = args.ds_ratios

    def get_job_url(self):
        # Split by "-master-" and take the first part
        master_addr = os.getenv("MASTER_ADDR", None)
        cluster = os.getenv("MLP_CLUSTER", None)

        def get_job_id(master_addr):
            if not master_addr:
                return None

            # Extract everything before the last "-master-" substring
            if "-master-" in master_addr:
                job_id = master_addr.split("-master-")[0]
                return job_id
            return master_addr
        if master_addr and cluster:
            job_id = get_job_id(master_addr)
            # https://va-mlp.anuttacon.com/console/dlc_job/job-yz8korktg6
            return f"https://{cluster}-mlp.anuttacon.com/console/dlc_job/{job_id}"
        return ""

    def trigger(self, model_path,  tflops, mfu, steps):
        """
        Trigger an evaluation by sending a POST request to the model management server.

        Args:
            model_path (str, optional): Override the model path. If None, uses args.model_path or checkpoint path
        """
        # Allow override of model_path and version_name
        assert model_path is not None, "model_path is required"
        world_size = dist.get_world_size()
        self.train_dp = world_size // (self.train_tp * self.train_pp)
        # Prepare the payload
        payload = {
            "project": self.project,
            "project_name": self.project_name,
            "modality": self.modality,
            "model_name": self.model_name,
            "cluster": self.cluster,
            "train_tp": self.train_tp,
            "train_pp": self.train_pp,
            "train_cp": self.train_cp,
            "train_dp": self.train_dp,
            "train_ep": self.train_ep,
            "train_sp": self.train_sp,
            "train_seqlen": self.train_seqlen,
            "train_batch_size": self.train_batch_size,
            "train_dataset_type": self.train_data_type,
            "train_dataset_name": self.train_dataset_name,
            "train_learning_rate": self.train_learning_rate,
            "train_tflops": tflops,
            "train_mfu": mfu,
            "train_tokenizer_path": self.train_tokenizer_path,
            "eval_tp": self.eval_tp,
            "eval_pp": self.eval_pp,
            "eval_cp": self.eval_cp,
            "eval_ep": self.eval_ep,
            "eval_sp": self.eval_sp, 
            "eval_seqlen": self.eval_seqlen,
            "eval_batch_size": self.eval_batch_size,
            "eval_ckpt_list": [model_path],
            "eval_dataset_list": self.eval_dataset_list,
            "eval_tasks": self.eval_tasks,
            "eval_tokenizer_path": self.eval_tokenizer_path if self.eval_tokenizer_path else self.train_tokenizer_path,
            "project": self.project,
            "train_steps": steps,
            "validation_batch_size": self.eval_validation_batch_size,
            "inference_max_seq_length": self.inference_max_seq_length,
            "code_info": self.code_info,
            "train_dataset_name_str": self.train_dataset_name_str,
            "train_dataset_ratios_str": self.train_dataset_ratios_str,
        }

        if os.getenv("WANDB_MODE") != "disabled":
            payload["wandb_project"] = self.wandb_project
            payload["wandb_api_key"] = self.wandb_api_key

        end_time = time.time()
        payload["train_job_url"] = self.get_job_url()
        payload["train_cost_time"] = round(
            (end_time - self.start_time) / 3600, 2)
        payload["finished_at"] = end_time

        return invoke_evaluation_service(payload, "evaluations")



import os
import io
import json
import mmap
import argparse
from typing import List, Tuple, Optional
from concurrent.futures import ProcessPoolExecutor, as_completed
from collections import defaultdict

os.environ.setdefault("TOKENIZERS_PARALLELISM", "false")

from transformers import AutoTokenizer

def make_line_aligned_chunks(path: str, chunk_bytes: int) -> List[Tuple[int, int]]:
    file_path = []
    # for single file
    if not os.path.isdir(path):
        file_path.append(path)
    else: 
        # for directory
        for entry in sorted(os.scandir(path), key=lambda e: e.name):
            if entry.is_file():
                file_path.append(entry.path)
    
    #file_to_range = defaultdict(list)
    chunks = defaultdict(list)
    for one_file in file_path:
        size = os.path.getsize(one_file)
        with open(one_file, "rb") as f:
            start = 0
            while start < size:
                end = min(start + chunk_bytes, size)
                if end < size:
                    f.seek(end)
                    f.readline()
                    end = f.tell()
                    if end <= start:  # 极端情况下保护（比如没读到换行）
                        end = min(start + chunk_bytes, size)
                #chunks.append((start, end))
                chunks[one_file].append((start, end))
                start = end
    return chunks

def iter_lines_mmap(path: str, start: int, end: int, encoding="utf-8"):
    with open(path, "rb") as f, mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mm:
        mm.seek(start)
        while mm.tell() < end:
            bline = mm.readline()
            if not bline:
                break
            pos = mm.tell()
            if pos > end:
                break
            yield bline.decode(encoding, errors="ignore").rstrip("\n")

def extract_text_from_jsonl_line(line: str) -> Optional[str]:
    try:
        obj = json.loads(line)
    except Exception:
        return None, 0

    invalid_count = 0
    if 'text' in obj:
        return obj['text'], 0
    elif 'messages' in obj:
        if not all(turn['role'] and turn['content'] and type(turn['role']) is str and type(turn['content']) is str for turn in obj['messages']):
            return None, 1
        text = ' '.join([f"{msg.get('role', '')} {msg.get('content', '')}" for msg in obj['messages']])
    else:
        print("invalid line:", line[:20] + "...")
        return None, 1

    return text, invalid_count

def count_tokens_batch(tokenizer, texts, add_special_tokens=False) -> int:
    if not texts:
        return 0
    enc = tokenizer(
        texts,
        add_special_tokens=add_special_tokens,
        return_attention_mask=False,
        return_token_type_ids=False,
    )
    return sum(len(ids) for ids in enc["input_ids"])

def worker_count_slice(
    path: str,
    start: int,
    end: int,
    tokenizer_path: str,
    batch_size: int,
    add_special_tokens: bool,
    rayon_threads: int,
    encoding: str,
) -> int:
    if rayon_threads > 0:
        os.environ["RAYON_NUM_THREADS"] = str(rayon_threads)

    tokenizer = AutoTokenizer.from_pretrained(tokenizer_path, use_fast=True)

    total = 0
    buf = []
    invalid_sample_count = 0
    for line in iter_lines_mmap(path, start, end, encoding=encoding):
        text, invalid = extract_text_from_jsonl_line(line)
        invalid_sample_count += invalid
        if not text:
            continue
        buf.append(text)
        if len(buf) >= batch_size:
            total += count_tokens_batch(tokenizer, buf, add_special_tokens)
            buf.clear()
    if buf:
        total += count_tokens_batch(tokenizer, buf, add_special_tokens)
        buf.clear()
    return total, invalid_sample_count

def count_tokens_jsonl_mmap_mp(
    path: str,
    tokenizer_path: str,
    workers: int = max(1, (os.cpu_count() or 8) // 2),
    chunk_bytes: int = 1 << 30,  # 1 GiB
    batch_size: int = 2048,
    add_special_tokens: bool = False,
    rayon_threads_per_proc: Optional[int] = None,
    encoding: str = "utf-8",
) -> int:
    chunks = make_line_aligned_chunks(path, chunk_bytes)
    chunk_num = sum(len(v) for v in chunks.values())
    print(f"Split to {chunk_num} chunks")
    if not chunks:
        return 0, 0
    workers = min(workers, chunk_num)

    cpu = os.cpu_count() or 8
    if rayon_threads_per_proc is None:
        rayon_threads_per_proc = max(1, cpu // max(1, workers))
    total_invalid_sample_count = 0
    total_tokens_count = 0
    with ProcessPoolExecutor(max_workers=workers) as ex:
        futs = []
        for file_path, chunk_list in chunks.items():
            for s, e in chunk_list:
                futs.append(
                    ex.submit(
                        worker_count_slice,
                        file_path, 
                        s, 
                        e,
                        tokenizer_path,
                        batch_size,
                        add_special_tokens,
                        rayon_threads_per_proc,
                        encoding,
                    )
                )
        for fut in as_completed(futs):
            tokens_count, invalid_sample_count = fut.result()
            total_tokens_count += tokens_count
            total_invalid_sample_count += invalid_sample_count
    return total_tokens_count, total_invalid_sample_count


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("--path", required=True, help="Path to the JSONL file")
    ap.add_argument("--tokenizer_path", required=True, help="Path/name to load with AutoTokenizer")
    ap.add_argument("--workers", type=int, default=max(1, (os.cpu_count() or 8)//2))
    ap.add_argument("--chunk_bytes", type=int, default=1<<30, help="Bytes per chunk (default: 1 GiB)")
    ap.add_argument("--batch_size", type=int, default=2048, help="Batch size for tokenization")
    ap.add_argument("--add_special_tokens", action="store_true", help="Include special tokens when counting")
    ap.add_argument("--rayon_threads_per_proc", type=int, default=None, help="Number of Rust (rayon) threads per process")
    ap.add_argument("--encoding", default="utf-8")
    ap.add_argument("--force", action='store_true', help="force re calculate iterations")
    ap.add_argument("--dataset_role", type=str, default="eval", choices=["train", "eval"])
    args = ap.parse_args()

    dataset_role_dir = os.path.join(args.path, args.dataset_role)
    if not os.path.isdir(dataset_role_dir):
        raise FileNotFoundError(f"Dataset directory {dataset_role_dir} does not exist")

    json_file = os.path.join(dataset_role_dir, "token_count.json")
    catched_data = {}
    if os.path.exists(json_file):
        with open(json_file, 'r') as f:
            catched_data = json.load(f)
        
    if (
        args.tokenizer_path in catched_data 
        and catched_data[args.tokenizer_path] != 0 
        and not args.force
    ):
        print(f"Token count for tokenizer {args.tokenizer_path} already exists: {catched_data[args.tokenizer_path]}")
        return

    total, total_invalid_count = count_tokens_jsonl_mmap_mp(
        path=dataset_role_dir,
        tokenizer_path=args.tokenizer_path,
        workers=args.workers,
        chunk_bytes=args.chunk_bytes,
        batch_size=args.batch_size,
        add_special_tokens=args.add_special_tokens,
        rayon_threads_per_proc=args.rayon_threads_per_proc,
        encoding=args.encoding,
    )
    print(f"total token is {total} but has {total_invalid_count} invalid samples")

    catched_data[args.tokenizer_path] = total
    catched_data["invalid_sample_count"] = total_invalid_count
    with open(json_file, "w", encoding="utf-8") as f:
        json.dump(catched_data, f, ensure_ascii=False, indent=2)

#python3 utils.py --path /mnt/project/llm/data/posttrain/foundation/v1.4_exp5 --dataset_role train --tokenizer_path /mnt/project/minghao/ckpts/CPT_72b/model_name=0__val_loss=1.88_step=103299_consumed_samples=16114800.0/nemo_to_hf
if __name__ == "__main__":
    import time
    t0 = time.time()
    main()
    elapsed_min = (time.time() - t0) / 60
    print(f"总耗时 {elapsed_min:.1f} 分钟")
